import subprocess
import tempfile
import venv

from loguru import logger


class EnvBuilder(venv.EnvBuilder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.context = None

    def post_setup(self, context):
        self.context = context


def run_in_different_venv(
    requirements_file: str,
    script_path: str,
    use_gpu: bool,
    *args,
):
    """Run a python scripts in a new temporary environment. Arguments for the
    script must be passed in the function args.
    it is equivalent to create and activate a new environment and running
    > pip install -r $requirement_file
    > python -m script_path *args
    Args:
        requirements_file (str): File (.txt) containing the list of
            requirements.
        script_path (str): Path to the script that must be run.
        args: Arguments of the script.
    """
    logger.debug(f"Debug: Running script {script_path} in a new virtual env.")
    with tempfile.TemporaryDirectory() as target_dir_path:
        logger.debug("Debug: Creating virtual environment...")
        venv_builder = EnvBuilder(with_pip=True)
        venv_builder.create(str(target_dir_path))
        venv_context = venv_builder.context

        logger.debug("Debug: Installing requirements...")

        if use_gpu:
            pip_install_command = [
                venv_context.env_exe,
                "-m",
                "pip",
                "install",
                "torch==1.9.1+cu111",
                "torchvision==0.10.1+cu111",
                "-f",
                "https://download.pytorch.org/whl/torch_stable.html",
            ]
        else:
            pip_install_command = [
                venv_context.env_exe,
                "-m",
                "pip",
                "install",
                "torch<=1.9.1",
                "torchvision<=0.10.1",
            ]
        subprocess.check_call(pip_install_command)

        pip_install_command = [
            venv_context.env_exe,
            "-m",
            "pip",
            "install",
            "-r",
            requirements_file,
        ]
        subprocess.check_call(pip_install_command)

        logger.debug("Debug: Executing script...")
        script_command = [venv_context.env_exe, script_path, *args]
        subprocess.check_call(script_command)



import tensorflow as tf
import torch

from nebullvm.tools.data import DataManager


def test_custom_input_data():
    input_data = [
        ((torch.randn(2, 3, 10, 10),), torch.randn(2, 1)),
        ((torch.randn(2, 3, 10, 10),), torch.randn(2, 1)),
        ((torch.randn(2, 3, 10, 10),), torch.randn(2, 1)),
        ((torch.randn(2, 3, 10, 10),), torch.randn(2, 1)),
    ]

    data_manager = DataManager(input_data)

    assert len(data_manager) == 4
    assert len(data_manager[0]) == 2
    assert len(data_manager[0][0]) == 1
    assert data_manager[0][0][0].shape == (2, 3, 10, 10)
    assert data_manager[0][1].shape == (2, 1)


def test_torch_dataloader_single_input_with_label():
    dataset = torch.utils.data.TensorDataset(
        torch.randn(8, 3, 10, 10), torch.randn(8, 1)
    )
    dataloader = torch.utils.data.DataLoader(dataset, batch_size=2)
    data_manager = DataManager.from_dataloader(dataloader)

    assert len(data_manager) == 4
    assert len(data_manager[0]) == 2
    assert len(data_manager[0][0]) == 1
    assert data_manager[0][0][0].shape == (2, 3, 10, 10)
    assert data_manager[0][1].shape == (2, 1)


def test_torch_dataloader_two_inputs_with_label():
    dataset = torch.utils.data.TensorDataset(
        torch.randn(8, 3, 10, 10), torch.randn(8, 3, 10, 10), torch.randn(8, 1)
    )

    dataloader = torch.utils.data.DataLoader(dataset, batch_size=2)
    data_manager = DataManager.from_dataloader(dataloader)

    assert len(data_manager) == 4
    assert len(data_manager[0]) == 2
    assert len(data_manager[0][0]) == 2
    assert data_manager[0][0][0].shape == (2, 3, 10, 10)
    assert data_manager[0][0][1].shape == (2, 3, 10, 10)
    assert data_manager[0][1].shape == (2, 1)


def test_torch_dataloader_three_inputs_with_label():
    dataset = torch.utils.data.TensorDataset(
        torch.randn(8, 3, 10, 10),
        torch.randn(8, 3, 10, 10),
        torch.randn(8, 3, 10, 10),
        torch.randn(8, 1),
    )
    dataloader = torch.utils.data.DataLoader(dataset, batch_size=2)
    data_manager = DataManager.from_dataloader(dataloader)

    assert len(data_manager) == 4
    assert len(data_manager[0]) == 2
    assert len(data_manager[0][0]) == 3
    assert data_manager[0][0][0].shape == (2, 3, 10, 10)
    assert data_manager[0][0][1].shape == (2, 3, 10, 10)
    assert data_manager[0][0][2].shape == (2, 3, 10, 10)
    assert data_manager[0][1].shape == (2, 1)


def test_torch_dataloader_single_input_without_label():
    dataset = torch.utils.data.TensorDataset(torch.randn(8, 3, 10, 10))
    dataloader = torch.utils.data.DataLoader(dataset, batch_size=2)
    data_manager = DataManager.from_dataloader(dataloader)

    assert len(data_manager) == 4
    assert len(data_manager[0]) == 2
    assert len(data_manager[0][0]) == 1
    assert data_manager[0][0][0].shape == (2, 3, 10, 10)


def test_tensorflow_dataloader_single_input_with_label():
    dataset = tf.data.Dataset.from_tensor_slices(
        (tf.random.normal([8, 10, 10, 3]), tf.random.normal([8, 1]))
    )
    data_manager = DataManager.from_dataloader(dataset.batch(2))

    assert len(data_manager) == 4
    assert len(data_manager[0]) == 2
    assert len(data_manager[0][0]) == 1
    assert data_manager[0][0][0].shape == (2, 10, 10, 3)
    assert data_manager[0][1].shape == (2, 1)


def test_tensorflow_dataloader_two_inputs_with_label():
    dataset = tf.data.Dataset.from_tensor_slices(
        (
            tf.random.normal([8, 10, 10, 3]),
            tf.random.normal([8, 10, 10, 3]),
            tf.random.normal([8, 1]),
        )
    )
    data_manager = DataManager.from_dataloader(dataset.batch(2))

    assert len(data_manager) == 4
    assert len(data_manager[0]) == 2
    assert len(data_manager[0][0]) == 2
    assert data_manager[0][0][0].shape == (2, 10, 10, 3)
    assert data_manager[0][0][1].shape == (2, 10, 10, 3)
    assert data_manager[0][1].shape == (2, 1)


def test_tensorflow_dataloader_three_inputs_with_label():
    dataset = tf.data.Dataset.from_tensor_slices(
        (
            tf.random.normal([8, 10, 10, 3]),
            tf.random.normal([8, 10, 10, 3]),
            tf.random.normal([8, 10, 10, 3]),
            tf.random.normal([8, 1]),
        )
    )
    data_manager = DataManager.from_dataloader(dataset.batch(2))

    assert len(data_manager) == 4
    assert len(data_manager[0]) == 2
    assert len(data_manager[0][0]) == 3
    assert data_manager[0][0][0].shape == (2, 10, 10, 3)
    assert data_manager[0][0][1].shape == (2, 10, 10, 3)
    assert data_manager[0][0][2].shape == (2, 10, 10, 3)
    assert data_manager[0][1].shape == (2, 1)


def test_tensorflow_dataloader_single_input_without_label():
    dataset = tf.data.Dataset.from_tensor_slices(
        tf.random.normal([8, 10, 10, 3])
    )
    data_manager = DataManager.from_dataloader(dataset.batch(2))

    assert len(data_manager) == 4
    assert len(data_manager[0]) == 2
    assert len(data_manager[0][0]) == 1
    assert data_manager[0][0][0].shape == (2, 10, 10, 3)



from pathlib import Path
from typing import List, Tuple, Optional, Dict, Union, Sequence

from nebullvm.optional_modules.torch import torch, Module, DataLoader
from nebullvm.tools.base import DataType, InputInfo, Device
from nebullvm.tools.data import DataManager

FX_MODULE_NAME = "NebullvmFxModule"


def save_with_torch_fx(model: Module, path: Path):
    traced_model = torch.fx.symbolic_trace(model)
    traced_model.to_folder(path, FX_MODULE_NAME)


def load_with_torch_fx(
    path: Path, state_dict_name: str = "pruned_state_dict.pt"
):
    module_file = path / "module.py"
    with open(module_file, "r") as f:
        module_str = f.read()
    exec(module_str, globals())
    model = eval(FX_MODULE_NAME)()
    model.load_state_dict(torch.load(path / state_dict_name))
    return model


def get_outputs_sizes_torch(
    torch_model: Module,
    input_tensors: List[torch.Tensor],
    device: Device,
) -> List[Tuple[int, ...]]:
    if device is Device.GPU:
        input_tensors = [x.cuda() for x in input_tensors]
        torch_model.cuda()
    with torch.no_grad():
        outputs = torch_model(*input_tensors)
        if isinstance(outputs, torch.Tensor):
            return [tuple(outputs.size())[1:]]
        else:
            return [tuple(output.size())[1:] for output in outputs]


def create_model_inputs_torch(
    batch_size: int, input_infos: List[InputInfo]
) -> List[torch.Tensor]:
    input_tensors = (
        torch.randn((batch_size, *input_info.size))
        if input_info.dtype is DataType.FLOAT32
        else torch.randint(
            size=(batch_size, *input_info.size),
            low=input_info.min_value or 0,
            high=input_info.max_value or 100,
        )
        for input_info in input_infos
    )
    return list(input_tensors)


def run_torch_model(
    torch_model: Module,
    input_tensors: List[torch.Tensor],
    device: Device,
    dtype: torch.dtype = torch.float,
) -> List[torch.Tensor]:
    torch_model.eval()
    if device is Device.GPU:
        torch_model.cuda()
        if dtype != torch.half:
            input_tensors = (t.cuda() for t in input_tensors)
        else:
            input_tensors = (
                t.cuda().half() if t.dtype == torch.float else t.cuda()
                for t in input_tensors
            )
    with torch.no_grad():
        pred = torch_model(*input_tensors)
    if isinstance(pred, torch.Tensor):
        pred = [pred.cpu()]
    else:
        pred = [p.cpu() for p in pred]
    return pred


def _extract_dynamic_axis(
    torch_model: Module,
    dataloader: DataManager,
    input_sizes: List[Tuple[int, ...]],
    batch_size: int,
    device: Device,
    max_data: int = 100,
) -> Optional[Dict]:
    from nebullvm.tools.utils import inspect_dynamic_size

    dynamic_axis = {"inputs": [{}] * len(input_sizes), "outputs": []}
    output_sizes = []
    for i, input_data in enumerate(dataloader):
        input_tensors = input_data[0]
        if i >= max_data:
            break
        inspect_dynamic_size(
            input_tensors, input_sizes, batch_size, dynamic_axis["inputs"]
        )
        outputs = tuple(run_torch_model(torch_model, input_tensors, device))
        if i == 0:
            dynamic_axis["outputs"] = [{}] * len(outputs)
            output_sizes = [tuple(output.shape[1:]) for output in outputs]
        inspect_dynamic_size(
            outputs, output_sizes, batch_size, dynamic_axis["outputs"]
        )
    if any(
        len(x) > 0 for x in (dynamic_axis["inputs"] + dynamic_axis["outputs"])
    ):
        return dynamic_axis
    return None


def extract_info_from_torch_data(
    model: Module,
    dataloader: Union[DataLoader, Sequence],
    batch_size: int,
    input_sizes: List[Tuple[int, ...]],
    input_types: List[str],
    dynamic_axis: Dict,
    device: Device,
):
    from nebullvm.tools.utils import ifnone

    input_data = (
        dataloader[0]
        if isinstance(dataloader, Sequence)
        else next(iter(dataloader))
    )
    input_row = input_data[0]

    batch_size = ifnone(batch_size, int(input_row[0].shape[0]))
    input_sizes = ifnone(input_sizes, [tuple(x.shape[1:]) for x in input_row])
    input_types = ifnone(
        input_types,
        [
            "int64"
            if isinstance(x.cpu(), torch.LongTensor)
            else "int32"
            if isinstance(x.cpu(), torch.IntTensor)
            else "float32"
            for x in input_row
        ],
    )

    if dynamic_axis is not None:
        dynamic_axis["inputs"] = [
            {int(k): v for (k, v) in val.items()}
            for val in dynamic_axis["inputs"]
        ]
        dynamic_axis["outputs"] = [
            {int(k): v for (k, v) in val.items()}
            for val in dynamic_axis["outputs"]
        ]

    dynamic_axis = ifnone(
        dynamic_axis,
        _extract_dynamic_axis(
            model, dataloader, input_sizes, batch_size, device
        ),
    )
    return batch_size, input_sizes, input_types, dynamic_axis


def torch_is_gpu_available():
    return torch.cuda.is_available()


def torch_get_device_name():
    return torch.cuda.get_device_name(0)



import abc
from pathlib import Path
from typing import Optional, List, Union

from nebullvm.operations.base import Operation
from nebullvm.operations.conversions.pytorch import convert_torch_to_onnx
from nebullvm.operations.conversions.tensorflow import convert_tf_to_onnx
from nebullvm.optional_modules.onnx import onnx
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.base import DeepLearningFramework, ModelParams
from nebullvm.tools.data import DataManager


class Converter(Operation, abc.ABC):
    ONNX_EXTENSION = ".onnx"
    TORCH_EXTENSION = ".pt"
    TF_EXTENSION = ".pb"

    def __init__(self, model_name: Optional[str] = None):
        super().__init__()
        self.model = None
        self.data = None
        self.converted_models = None
        self.model_params = None
        self.device = None
        self.model_name = model_name or "temp"

    def set_state(
        self, model: Union[torch.nn.Module, tf.Module, str], data: DataManager
    ):
        self.model = model
        self.data = data
        return self

    def get_result(self) -> List:
        return self.converted_models


class PytorchConverter(Converter):
    DEST_FRAMEWORKS = [DeepLearningFramework.NUMPY]

    def execute(
        self,
        save_path: Path,
        model_params: ModelParams,
    ):
        self.converted_models = [self.model]
        for framework in self.DEST_FRAMEWORKS:
            if framework is DeepLearningFramework.NUMPY:
                self.onnx_conversion(save_path, model_params)
            else:
                raise NotImplementedError()

    def onnx_conversion(self, save_path, model_params):
        onnx_path = save_path / f"{self.model_name}{self.ONNX_EXTENSION}"
        onnx_model_path = convert_torch_to_onnx(
            torch_model=self.model,
            input_data=self.data,
            model_params=model_params,
            output_file_path=onnx_path,
            device=self.device,
        )
        if self.converted_models is None:
            self.converted_models = [onnx_model_path]
        else:
            self.converted_models.append(onnx_model_path)

    def tensorflow_conversion(self):
        # TODO: Implement conversion from Pytorch to Tensorflow
        raise NotImplementedError()


class TensorflowConverter(Converter):
    DEST_FRAMEWORKS = [DeepLearningFramework.NUMPY]

    def execute(
        self,
        save_path: Path,
        model_params: ModelParams,
    ):
        self.converted_models = [self.model]
        for framework in self.DEST_FRAMEWORKS:
            if framework is DeepLearningFramework.NUMPY:
                self.onnx_conversion(save_path, model_params)
            else:
                raise NotImplementedError()

    def onnx_conversion(self, save_path, model_params):
        onnx_path = save_path / f"{self.model_name}{self.ONNX_EXTENSION}"
        onnx_model_path = convert_tf_to_onnx(
            model=self.model,
            model_params=model_params,
            output_file_path=onnx_path,
        )
        if self.converted_models is None:
            self.converted_models = [onnx_model_path]
        else:
            self.converted_models.append(onnx_model_path)

    def pytorch_conversion(self):
        # TODO: Implement conversion from Tensorflow to Pytorch
        raise NotImplementedError()


class ONNXConverter(Converter):
    DEST_FRAMEWORKS = []

    def execute(self, save_path, model_params):
        onnx_path = save_path / f"{self.model_name}{self.ONNX_EXTENSION}"
        try:
            model_onnx = onnx.load(str(self.model))
            onnx.save(model_onnx, str(onnx_path))
        except Exception:
            self.logger.error(
                "The provided onnx model path is invalid. Please provide"
                " a valid path to a model in order to use Nebullvm."
            )
            self.converted_models = []

        self.converted_models = [str(onnx_path)]

    def tensorflow_conversion(self):
        # TODO: Implement conversion from ONNX to Tensorflow
        raise NotImplementedError()

    def pytorch_conversion(self):
        # TODO: Implement conversion from ONNX to Pytorch
        raise NotImplementedError()



from pathlib import Path
from typing import Union, Iterable, Tuple, List

import cpuinfo
import numpy as np

from nebullvm.optional_modules.onnx import (
    onnx,
    convert_float_to_float16_model_path,
)
from nebullvm.optional_modules.onnxruntime import (
    CalibrationDataReader,
    QuantType,
    quantize_dynamic,
    quantize_static,
)
from nebullvm.optional_modules.torch import DataLoader
from nebullvm.tools.base import QuantizationType, Device
from nebullvm.tools.onnx import get_input_names
from nebullvm.tools.transformations import (
    MultiStageTransformation,
    HalfPrecisionTransformation,
)


class _IterableCalibrationDataReader(CalibrationDataReader):
    def __init__(
        self,
        iterable_dataset: Union[Iterable[Tuple], List[Tuple]],
        input_names: List[str],
    ):
        self.iterable_dataset = iter(
            [
                {
                    input_name: value
                    for inputs in iterable_dataset
                    for input_name, value in zip(input_names, inputs)
                }
            ]
        )

    def get_next(self) -> dict:
        return next(self.iterable_dataset, None)

    @classmethod
    def from_dataloader(
        cls, dl: DataLoader, input_names: List[str], contains_y: bool = True
    ):
        iterable_ds = iter(
            inputs[:-1] if contains_y else inputs for inputs in dl
        )
        return cls(iterable_ds, input_names)


def _quantize_dynamic(model_path: str) -> str:
    model_path = Path(model_path)
    model_quant = model_path.parent.parent / "int8_dynamic"
    model_quant.mkdir(parents=True, exist_ok=True)
    model_quant = model_quant / (model_path.stem + ".quant.onnx")
    quantize_dynamic(
        model_path,
        model_quant,
        weight_type=QuantType.QUInt8,
        optimize_model=False,
    )
    return str(model_quant)


def _get_quantization_type_for_static(use_gpu) -> Tuple[QuantType, QuantType]:
    """Returns the quantization types for activations and weights,
    depending on the underlying hardware
    """
    arch = cpuinfo.get_cpu_info()["arch"].lower()
    if use_gpu:
        activation_type = weight_type = QuantType.QInt8
    elif "x86" in arch:
        cpu_raw_data = cpuinfo.get_cpu_info()["brand_raw"].lower()
        if "intel" in cpu_raw_data and "xeon" in cpu_raw_data:
            activation_type = QuantType.QUInt8
            weight_type = QuantType.QInt8
        else:
            activation_type = weight_type = QuantType.QUInt8
    else:
        activation_type = QuantType.QUInt8
        weight_type = QuantType.QUInt8
    return activation_type, weight_type


def _quantize_static(
    model_path: str, input_data: List[Tuple[np.ndarray, ...]], use_gpu: bool
) -> str:
    model_path = Path(model_path)
    model_quant = model_path.parent.parent / "int8_static"
    model_quant.mkdir(parents=True, exist_ok=True)
    model_quant = model_quant / (model_path.stem + ".quant.onnx")
    inputs = input_data
    input_names = get_input_names(str(model_path))
    cdr = _IterableCalibrationDataReader(
        input_names=input_names, iterable_dataset=inputs
    )
    activation_type, weight_type = _get_quantization_type_for_static(use_gpu)
    quantize_static(
        model_path,
        Path(model_quant),
        cdr,
        activation_type=activation_type,
        weight_type=weight_type,
        optimize_model=False,
    )
    return str(model_quant)


def _convert_to_half_precision(
    model_path: str, input_tfms: MultiStageTransformation
) -> str:
    model_path = Path(model_path)
    model_quant = model_path.parent.parent / "fp16"
    model_quant.mkdir(parents=True)
    model_quant = model_quant / (model_path.stem + "_fp16.onnx")
    new_onnx_model = convert_float_to_float16_model_path(str(model_path))
    input_tfms.append(HalfPrecisionTransformation())
    onnx.save(new_onnx_model, str(model_quant))
    return str(model_quant)


def quantize_onnx(
    model_path: str,
    input_data: List[Tuple[np.ndarray, ...]],
    quantization_type: QuantizationType,
    device: Device,
    input_tfms: MultiStageTransformation,
) -> str:
    if quantization_type == QuantizationType.DYNAMIC:
        return _quantize_dynamic(model_path)
    elif quantization_type == QuantizationType.STATIC:
        return _quantize_static(model_path, input_data, device is Device.GPU)
    elif quantization_type == QuantizationType.HALF:
        return _convert_to_half_precision(model_path, input_tfms)
    else:
        raise ValueError(
            f"Quantization type {quantization_type} not supported"
        )



from typing import Union, List, Tuple

from nebullvm.config import QUANTIZATION_DATA_NUM
from nebullvm.operations.optimizations.compilers.base import Compiler

from nebullvm.operations.optimizations.compilers.quantizations.pytorch import (
    quantize_pytorch,
)
from nebullvm.operations.optimizations.compilers.quantizations.utils import (
    check_quantization,
)
from nebullvm.optional_modules.torch import (
    torch,
    Module,
    ScriptModule,
    GraphModule,
    symbolic_trace,
)
from nebullvm.tools.base import QuantizationType, Device
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import MultiStageTransformation


class PytorchBackendCompiler(Compiler):
    supported_ops = {
        "cpu": [None, QuantizationType.STATIC, QuantizationType.DYNAMIC],
        "gpu": [
            None,
            QuantizationType.HALF,
        ],
    }

    def execute(
        self,
        model: Module,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Optimize the input model using pytorch built-in techniques.

        Args:
            model (torch.nn.Module): The pytorch model.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction. Default: None.
            metric_drop_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with a higher drop
                will be ignored. Default: None.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used. Default: None.
            input_data (DataManager): User defined data. Default: None.
        """

        if quantization_type not in self.supported_ops[self.device.value]:
            self.compiled_model = None
            return

        if quantization_type is QuantizationType.STATIC and input_data is None:
            raise ValueError("Input data is required for static quantization.")

        self.logger.info(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )

        check_quantization(quantization_type, metric_drop_ths)
        train_input_data = input_data.get_split("train").get_list(
            QUANTIZATION_DATA_NUM
        )

        if quantization_type is not None:
            model = self._quantize_model(
                model, quantization_type, input_tfms, train_input_data
            )

        self.compiled_model = self._compile_model(
            model, input_data, quantization_type
        )

    def _compile_model(
        self,
        model: Union[Module, GraphModule],
        input_data: DataManager,
        quantization_type: QuantizationType,
    ) -> ScriptModule:
        input_sample = input_data.get_list(1)[0]
        if self.device is Device.GPU:
            if quantization_type is QuantizationType.HALF:
                input_sample = [t.cuda().half() for t in input_sample]
            else:
                input_sample = [t.cuda() for t in input_sample]

        if not isinstance(model, torch.fx.GraphModule):
            model.eval()
            try:
                model_scripted = symbolic_trace(model)
                model_scripted = torch.jit.script(model_scripted)
            except Exception:
                try:
                    model_scripted = torch.jit.script(model)
                except Exception:
                    model_scripted = torch.jit.trace(model, input_sample)
        else:
            model_scripted = torch.jit.script(model)

        return model_scripted

    def _quantize_model(
        self,
        model: Module,
        quantization_type: QuantizationType,
        input_tfms: MultiStageTransformation,
        input_data_torch: List[Tuple[torch.Tensor, ...]],
    ):
        return quantize_pytorch(
            model, quantization_type, input_tfms, input_data_torch, self.device
        )



from loguru import logger

from nebullvm.tools.base import QuantizationType


def check_quantization(
    quantization_type: QuantizationType, perf_loss_ths: float
):
    if quantization_type is not None and perf_loss_ths is None:
        logger.warning(
            "Got a valid quantization type without any given quantization "
            "threshold. The quantization step will be ignored."
        )



import subprocess
import sys
from pathlib import Path

import numpy as np
from packaging import version
from types import ModuleType
from typing import (
    Tuple,
    Any,
    List,
    Dict,
    Union,
    Iterable,
    Sequence,
    Optional,
    Callable,
)

from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.base import DeepLearningFramework, Device, ModelParams
from nebullvm.tools.data import DataManager
from nebullvm.tools.onnx import (
    extract_info_from_np_data,
    get_output_sizes_onnx,
)
from nebullvm.tools.pytorch import (
    extract_info_from_torch_data,
    get_outputs_sizes_torch,
)
from nebullvm.tools.tf import (
    extract_info_from_tf_data,
    get_outputs_sizes_tf,
)


def ifnone(target, new_value):
    if target is None:
        return new_value
    else:
        return target


def inspect_dynamic_size(
    tensors: Tuple[Any, ...],
    sizes: List[Tuple[int, ...]],
    batch_size: int,
    axis_list: List[Dict],
):
    for idx, (tensor, size) in enumerate(zip(tensors, sizes)):
        size = (batch_size, *size)
        for idy, (j, k) in enumerate(zip(tensor.shape, size)):
            if j != k:
                if idy == 0:
                    tag = "batch_size"
                else:
                    tag = f"val_{j}_{k}"
                axis_list[idx][idy] = tag


def gpu_is_available():
    try:
        subprocess.check_output("nvidia-smi")
        return True
    except Exception:
        return False


def check_module_version(
    module: ModuleType, min_version: str = None, max_version: str = None
) -> bool:
    installed_version = module.__version__

    if min_version is not None:
        if version.parse(installed_version) < version.parse(min_version):
            return False

    if max_version is not None:
        if version.parse(installed_version) > version.parse(max_version):
            return False

    return True


def is_python_version_3_10():
    return (
        str(sys.version_info.major) + "." + str(sys.version_info.minor)
        == "3.10"
    )


def get_dl_framework(model: Any):
    if isinstance(model, torch.nn.Module):
        return DeepLearningFramework.PYTORCH
    elif isinstance(model, tf.Module) and model is not None:
        return DeepLearningFramework.TENSORFLOW
    elif isinstance(model, str):
        if Path(model).is_file():
            return DeepLearningFramework.NUMPY
        else:
            raise FileNotFoundError(
                f"No file '{model}' found, please provide a valid path to "
                f"a model."
            )
    else:
        raise TypeError(f"Model type {type(model)} not supported.")


def check_input_data(input_data: Union[Iterable, Sequence]):
    try:
        assert len(input_data) > 0
        assert isinstance(input_data[0], tuple)
        assert isinstance(input_data[0][0], tuple)
        assert isinstance(
            input_data[0][0][0], (np.ndarray, torch.Tensor, tf.Tensor)
        )
        assert isinstance(
            input_data[0][1],
            (np.ndarray, torch.Tensor, tf.Tensor, int, float, type(None)),
        )
    except:  # noqa E722
        return False
    else:
        return True


def is_data_subscriptable(input_data: Union[Iterable, Sequence]):
    try:
        input_data[0]
    except:  # noqa E722
        return False
    else:
        return True


def extract_info_from_data(
    model: Any,
    input_data: DataManager,
    dl_framework: DeepLearningFramework,
    dynamic_info: Optional[Dict],
    device: Device,
):
    batch_size, input_sizes, input_types, dynamic_info = INFO_EXTRACTION_DICT[
        dl_framework
    ](
        model,
        input_data,
        batch_size=None,
        input_sizes=None,
        input_types=None,
        dynamic_axis=dynamic_info,
        device=device,
    )
    model_params = ModelParams(
        batch_size=batch_size,
        input_infos=[
            {"size": size, "dtype": dtype}
            for size, dtype in zip(input_sizes, input_types)
        ],
        output_sizes=OUTPUT_SIZE_COMPUTATION_DICT[dl_framework](
            model, input_data[0][0], device
        ),
        dynamic_info=dynamic_info,
    )
    return model_params


def is_huggingface_data(data_sample: Any) -> bool:
    if is_dict_type(data_sample):
        return True
    elif isinstance(data_sample, str):
        return True
    elif isinstance(data_sample[0], str):
        return True
    return False


def is_dict_type(data_sample: Any):
    try:
        data_sample.items()
    except AttributeError:
        return False
    else:
        return True


INFO_EXTRACTION_DICT: Dict[DeepLearningFramework, Callable] = {
    DeepLearningFramework.PYTORCH: extract_info_from_torch_data,
    DeepLearningFramework.TENSORFLOW: extract_info_from_tf_data,
    DeepLearningFramework.NUMPY: extract_info_from_np_data,
}

OUTPUT_SIZE_COMPUTATION_DICT: Dict[DeepLearningFramework, Callable] = {
    DeepLearningFramework.PYTORCH: get_outputs_sizes_torch,
    DeepLearningFramework.TENSORFLOW: get_outputs_sizes_tf,
    DeepLearningFramework.NUMPY: get_output_sizes_onnx,
}



from collections import OrderedDict
from typing import List, Any, Dict

from nebullvm.operations.inference_learners.base import (
    InferenceLearnerWrapper,
    PytorchBaseInferenceLearner,
    LearnerMetadata,
)
from nebullvm.tools.huggingface import restructure_output


class HuggingFaceInferenceLearner(InferenceLearnerWrapper):
    """Class wrapping an InferenceLearner model and giving to it the
    huggingface interface.

    The class fuse both the InterfaceLearner and HuggingFace interfaces, giving
    to the final user a model which can be used whit the prefered API without
    the need of adapting the previous code.

    Attributes:
        network_parameters (ModelParams): Model parameters of the model.
        core_inference_learner (PytorchBaseInferenceLearner): Inference learner
            built using the Pytorch interface.
        output_structure (Dict): Original output structure of the HuggingFace
            model.
        input_names (List[str]): List of all the input keys used for the
            original HuggingFace model.
        output_type (Any, optional): Original output type of the HuggingFace
            model.
    """

    def __init__(
        self,
        core_inference_learner: PytorchBaseInferenceLearner,
        output_structure: OrderedDict,
        input_names: List[str],
        output_type: Any = None,
    ):
        super().__init__(core_inference_learner)
        self.output_structure = output_structure
        self.input_names = input_names
        self.output_type = output_type

    def _save_wrapper_extra_info(self):
        pass

    def get_size(self):
        return self.core_inference_learner.get_size()

    @staticmethod
    def _load_wrapper_extra_info(builder_inputs: Dict) -> Dict:
        return builder_inputs

    def run(self, *args, **kwargs) -> Any:
        """Run the underlying optimized model for getting a prediction.

        The method has an hybrid interface. It accepts inputs either as
        positional or keyword arguments. If only positional arguments are given
        the method expects the inputs to be in the canonical
        nebullvm interface. If only keyword arguments are given the method
        expects them to be in the HuggingFace interface. Mixed representation
        is not allowed and will result in an error.
        """
        if len(args) > 0 and len(kwargs) > 0:
            raise RuntimeError(
                "Not allowed usage of the predict method. "
                "Either the positional or the keyword arguments must be given."
            )
        if len(args) > 0:
            return self.core_inference_learner(*args)
        inputs = (kwargs.pop(name) for name in self.input_names)
        outputs = self.core_inference_learner(*inputs)

        if self.output_type is tuple:
            return outputs
        else:
            return restructure_output(
                outputs, self.output_structure, self.output_type
            )

    def _get_extra_metadata_kwargs(self) -> Dict:
        metadata_kwargs = {
            "output_structure": self.output_structure,
            "output_structure_keys": list(self.output_structure.keys()),
            "input_names": self.input_names,
        }
        if self.output_type is not None:
            metadata_kwargs.update(
                {
                    "output_type": self.output_type.__name__,
                    "output_type_module": self.output_type.__module__,
                }
            )
        return metadata_kwargs

    @staticmethod
    def _convert_metadata_to_inputs(metadata: LearnerMetadata) -> Dict:
        # we need to guarantee the preservation of the output structure
        # elements order.
        output_structure = OrderedDict()
        for key in metadata["output_structure_keys"]:
            output_structure[key] = metadata["output_structure"][key]

        inputs = {
            "output_structure": output_structure,
            "input_names": metadata["input_names"],
        }
        if metadata["output_type"] is not None:
            exec(
                f"from {metadata['output_type_module']} "
                f"import {metadata['output_type']}"
            )
            inputs["output_type"] = eval(metadata["output_type"])
        return inputs



import pickle
from abc import ABC
from pathlib import Path
from typing import Union, Tuple, Dict, Type

from loguru import logger

from nebullvm.operations.inference_learners.base import (
    BaseInferenceLearner,
    LearnerMetadata,
    PytorchBaseInferenceLearner,
)
from nebullvm.optional_modules.neural_compressor import (
    cfgs_to_fx_cfgs,
    cfg_to_qconfig,
)
from nebullvm.optional_modules.torch import (
    torch,
    prepare_fx,
    convert_fx,
    Module,
    GraphModule,
)
from nebullvm.tools.base import ModelParams, DeepLearningFramework
from nebullvm.tools.pytorch import (
    save_with_torch_fx,
    load_with_torch_fx,
    create_model_inputs_torch,
)
from nebullvm.tools.transformations import MultiStageTransformation
from nebullvm.tools.utils import check_module_version


class NeuralCompressorInferenceLearner(BaseInferenceLearner, ABC):
    """Model optimized on CPU using IntelNeuralCompressor.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        model (torch.fx.GraphModule): Torch fx graph model.
    """

    name = "IntelNeuralCompressor"

    def __init__(
        self,
        model: Union[Module, GraphModule],
        model_quant: GraphModule,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.model = model
        self.model_quant = model_quant

    def get_size(self):
        return len(pickle.dumps(self.model_quant, -1)) + len(
            pickle.dumps(self.model, -1)
        )

    def save(self, path: Union[str, Path], **kwargs):
        """Save the model.

        Args:
            path (Path or str): Path to the directory where the model will
                be stored.
            kwargs (Dict): Dictionary of key-value pairs that will be saved in
                the model metadata file.
        """
        metadata = LearnerMetadata.from_model(self, **kwargs)
        metadata.save(path)

        path_orig_model = Path(path) / Path("model_orig")
        path_quant_model = Path(path) / Path("model_quant")

        save_with_torch_fx(self.model, path_orig_model)
        self.model_quant.save(str(path_quant_model))

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        """Load the model.

        Args:
            path (Path or str): Path to the directory where the model is
                stored.
            kwargs (Dict): Dictionary of additional arguments for consistency
                with other Learners.

        Returns:
            DeepSparseInferenceLearner: The optimized model.
        """
        if len(kwargs) > 0:
            logger.warning(
                f"No extra keywords expected for the load method. "
                f"Got {kwargs}."
            )

        metadata = LearnerMetadata.read(path)
        input_tfms = metadata.input_tfms
        if input_tfms is not None:
            input_tfms = MultiStageTransformation.from_dict(
                metadata.input_tfms
            )

        network_parameters = ModelParams(**metadata.network_parameters)

        path_orig_model = Path(path) / Path("model_orig")
        path_quant_model = Path(path) / Path("model_quant") / "best_model.pt"

        model = load_with_torch_fx(
            Path(path_orig_model), "state_dict.pt"
        ).eval()
        state_dict = torch.load(path_quant_model)

        tune_cfg = state_dict.pop("best_configure")
        op_cfgs = cfg_to_qconfig(tune_cfg, tune_cfg["approach"])
        fx_op_cfgs = cfgs_to_fx_cfgs(op_cfgs, tune_cfg["approach"])

        additional_arguments = {}
        if check_module_version(torch, min_version="1.13.0"):
            additional_arguments["example_inputs"] = tuple(
                create_model_inputs_torch(
                    batch_size=network_parameters.batch_size,
                    input_infos=network_parameters.input_infos,
                )
            )

        q_model = prepare_fx(
            model,
            fx_op_cfgs,
            **additional_arguments,
        )
        q_model = convert_fx(q_model)

        q_model.load_state_dict(state_dict)

        return cls(
            input_tfms=input_tfms,
            network_parameters=ModelParams(**metadata.network_parameters),
            model=model,
            model_quant=q_model,
        )


class PytorchNeuralCompressorInferenceLearner(
    NeuralCompressorInferenceLearner, PytorchBaseInferenceLearner
):
    """Model optimized on CPU using IntelNeuralCompressor.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        model (torch.fx.GraphModule): Torch fx graph model.
    """

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        inputs = (t.cpu() for t in input_tensors)
        outputs = self.model_quant(*inputs)

        if isinstance(outputs, torch.Tensor):
            outputs = (outputs,)

        return outputs


NEURAL_COMPRESSOR_INFERENCE_LEARNERS: Dict[
    DeepLearningFramework, Type[NeuralCompressorInferenceLearner]
] = {DeepLearningFramework.PYTORCH: PytorchNeuralCompressorInferenceLearner}



import pickle
from pathlib import Path
from typing import Tuple, Union, Dict, Type

from nebullvm.config import TENSORFLOW_BACKEND_FILENAMES
from nebullvm.operations.inference_learners.base import (
    TensorflowBaseInferenceLearner,
    LearnerMetadata,
)
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.tools.base import Device, ModelParams


class TensorflowBackendInferenceLearner(TensorflowBaseInferenceLearner):
    name = "XLA"

    def __init__(self, tf_model: tf.Module, device: Device, **kwargs):
        super(TensorflowBackendInferenceLearner, self).__init__(**kwargs)
        self.model = tf_model
        self.device = device

    def get_size(self):
        return len(pickle.dumps(self.model, -1))

    def run(self, *input_tensors: tf.Tensor) -> Tuple[tf.Tensor, ...]:
        with tf.device(self.device.value):
            res = self.model(input_tensors)
        if not isinstance(res, tuple):
            return (res,)
        return res

    def save(self, path: Union[str, Path], **kwargs):
        path = Path(path)
        path.mkdir(exist_ok=True)
        metadata = LearnerMetadata.from_model(self, **kwargs)
        metadata.save(path)
        self.model.save(path / TENSORFLOW_BACKEND_FILENAMES["tf_model"])

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        path = Path(path)
        metadata = LearnerMetadata.read(path)
        network_parameters = ModelParams(**metadata.network_parameters)
        input_tfms = metadata.input_tfms
        model = tf.keras.models.load_model(
            path / TENSORFLOW_BACKEND_FILENAMES["tf_model"]
        )
        device = metadata.device
        return cls(
            tf_model=model,
            network_parameters=network_parameters,
            input_tfms=input_tfms,
            device=device,
        )


class TFLiteBackendInferenceLearner(TensorflowBaseInferenceLearner):
    name = "TFLite"

    def __init__(self, tflite_file: bytes, device: Device, **kwargs):
        super(TFLiteBackendInferenceLearner, self).__init__(**kwargs)
        self.tflite_file = tflite_file
        self.interpreter = tf.lite.Interpreter(model_content=tflite_file)
        self.device = device

    def get_size(self):
        return len(self.tflite_file)

    def run(self, *input_tensors: tf.Tensor):
        input_details = self.interpreter.get_input_details()
        output_details = self.interpreter.get_output_details()
        if self.network_parameters.dynamic_info:
            for i, (input_tensor, detail) in enumerate(
                zip(input_tensors, input_details)
            ):
                if input_tensor.shape != tuple(detail["shape"]):
                    self.interpreter.resize_tensor_input(i, input_tensor.shape)
        self.interpreter.allocate_tensors()
        for i, input_tensor in enumerate(input_tensors):
            self.interpreter.set_tensor(i, input_tensor)
        self.interpreter.invoke()
        return tuple(
            tf.convert_to_tensor(
                self.interpreter.get_tensor(output_detail["index"])
            )
            for output_detail in output_details
        )

    def save(self, path: Union[str, Path], **kwargs):
        path = Path(path)
        metadata = LearnerMetadata.from_model(self, **kwargs)
        metadata.save(path)
        with open(
            path / TENSORFLOW_BACKEND_FILENAMES["tflite_model"], "wb"
        ) as f:
            f.write(self.tflite_file)

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        path = Path(path)
        tflite_file_path = str(
            path / TENSORFLOW_BACKEND_FILENAMES["tflite_model"]
        )

        with open(tflite_file_path, "rb") as f:
            tflite_file = f.read()

        metadata = LearnerMetadata.read(path)
        network_parameters = ModelParams(**metadata.network_parameters)
        input_tfms = metadata.input_tfms
        device = metadata.device
        return cls(
            tflite_file=tflite_file,
            network_parameters=network_parameters,
            input_tfms=input_tfms,
            device=device,
        )


TF_BACKEND_LEARNERS_DICT: Dict[
    str,
    Type[
        Union[TensorflowBackendInferenceLearner, TFLiteBackendInferenceLearner]
    ],
] = {
    "tf": TensorflowBackendInferenceLearner,
    "tflite": TFLiteBackendInferenceLearner,
}



import json
import shutil
from abc import ABC
from pathlib import Path
from typing import Dict, Union, Type, Generator, Tuple, List, Optional

import numpy as np
from loguru import logger

from nebullvm.config import OPENVINO_FILENAMES
from nebullvm.operations.inference_learners.base import (
    BaseInferenceLearner,
    LearnerMetadata,
    PytorchBaseInferenceLearner,
    TensorflowBaseInferenceLearner,
    NumpyBaseInferenceLearner,
)
from nebullvm.optional_modules.openvino import (
    Core,
    Model,
    CompiledModel,
    InferRequest,
)
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.base import ModelParams, DeepLearningFramework
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import MultiStageTransformation


class OpenVinoInferenceLearner(BaseInferenceLearner, ABC):
    """Model optimized using OpenVINO.

    The class cannot be directly instantiated, but implements all the core
    methods needed for using OpenVINO at inference time.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        exec_network (any): The graph executor. This is the
            central component in the OpenVino optimized model execution.
        input_keys (List): Keys associated to the inputs.
        output_keys (List): Keys associated to the outputs.
        description_file (str): File containing a description of the optimized
            model.
        weights_file (str): File containing the model weights.
    """

    MODEL_NAME = "model.bin"
    name = "OpenVINO"

    def __init__(
        self,
        compiled_model: CompiledModel,
        infer_request: InferRequest,
        input_keys: List,
        output_keys: List,
        description_file: str,
        weights_file: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.compiled_model = compiled_model
        self.infer_request = infer_request
        self.input_keys = input_keys
        self.output_keys = output_keys
        self.description_file = self._store_file(description_file)
        self.weights_file = self._store_file(weights_file)

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        """Load the model.

        Args:
            path (Path or str): Path to the directory where the model is
                stored.
            kwargs (Dict): Dictionary of additional arguments for the
                `from_model_name` class method.

        Returns:
            OpenVinoInferenceLearner: The optimized model.
        """
        path = Path(path)

        with open(path / OPENVINO_FILENAMES["metadata"], "r") as fin:
            metadata = json.load(fin)
        metadata.update(kwargs)
        metadata["network_parameters"] = ModelParams(
            **metadata["network_parameters"]
        )
        input_tfms = metadata.get("input_tfms")
        if input_tfms is not None:
            metadata["input_tfms"] = MultiStageTransformation.from_dict(
                input_tfms
            )

        model_name = str(path / OPENVINO_FILENAMES["description_file"])
        model_weights = str(path / OPENVINO_FILENAMES["weights"])
        return cls.from_model_name(
            model_name=model_name, model_weights=model_weights, **metadata
        )

    def get_size(self):
        return len(self.compiled_model.export_model())

    @classmethod
    def from_model_name(
        cls,
        network_parameters: ModelParams,
        model_name: str,
        model_weights: str,
        input_tfms: MultiStageTransformation = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Build the optimized model from the network description and its
        weights.

        Args:
            network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
            model_name (str): File containing a description of the optimized
                model.
            model_weights (str): File containing the model weights.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            input_data (DataManager, optional): User defined data.
        """
        if len(kwargs) > 0:
            logger.warning(f"Found extra parameters: {kwargs}")

        core = Core()
        model = core.read_model(model=model_name, weights=model_weights)

        dynamic_shape = cls._get_dynamic_shape(model, network_parameters)

        if dynamic_shape is not None:
            model.reshape(dynamic_shape)

        compiled_model = core.compile_model(model=model, device_name="CPU")
        infer_request = compiled_model.create_infer_request()

        input_keys = list(
            map(lambda obj: obj.get_any_name(), compiled_model.inputs)
        )
        output_keys = list(
            map(lambda obj: obj.get_any_name(), compiled_model.outputs)
        )

        return cls(
            compiled_model,
            infer_request,
            input_keys,
            output_keys,
            input_tfms=input_tfms,
            network_parameters=network_parameters,
            description_file=model_name,
            weights_file=model_weights,
            input_data=input_data,
        )

    @staticmethod
    def _get_dynamic_shape(
        model: Model, network_parameters: ModelParams
    ) -> Optional[Dict[str, Tuple[int]]]:
        if network_parameters.dynamic_info is None:
            return None

        input_names = [
            list(model_input.names)[0] for model_input in model.inputs
        ]
        input_shapes = [
            (network_parameters.batch_size, *input_info.size)
            for input_info in network_parameters.input_infos
        ]
        dynamic_shapes = []

        assert len(input_shapes) == len(
            network_parameters.dynamic_info.inputs
        ), (
            f"Number of inputs defined in dynamic info "
            f"({len(input_shapes)}) is different from the one "
            f"expected from the model "
            f"({len(network_parameters.dynamic_info.inputs)})."
        )

        for input_shape, dynamic_shape_dict in zip(
            input_shapes, network_parameters.dynamic_info.inputs
        ):
            input_shape = list(input_shape)
            for key in dynamic_shape_dict.keys():
                input_shape[int(key)] = -1
            dynamic_shapes.append(tuple(input_shape))

        dynamic_shape_dict = {
            k: v for k, v in zip(input_names, dynamic_shapes)
        }
        return dynamic_shape_dict

    def _get_metadata(self, **kwargs) -> LearnerMetadata:
        # metadata = {
        #     key: self.__dict__[key] for key in ("input_keys", "output_keys")
        # }
        metadata = {}
        metadata.update(kwargs)
        return LearnerMetadata.from_model(self, **metadata)

    def save(self, path: Union[str, Path], **kwargs):
        """Save the model.

        Args:
            path (Path or str): Path to the directory where the model will
                be stored.
            kwargs (Dict): Dictionary of key-value pairs that will be saved in
                the model metadata file.
        """
        path = Path(path)
        path.mkdir(exist_ok=True)
        metadata = self._get_metadata(**kwargs)

        metadata.save(path)

        shutil.copy(
            self.description_file,
            path / OPENVINO_FILENAMES["description_file"],
        )
        shutil.copy(self.weights_file, path / OPENVINO_FILENAMES["weights"])

    def _predict_array(
        self,
        input_arrays: Generator[np.ndarray, None, None],
    ) -> Generator[np.ndarray, None, None]:

        results = self.infer_request.infer(
            inputs={
                input_key: input_array
                for input_key, input_array in zip(
                    self.input_keys, input_arrays
                )
            }
        )
        results = {
            output_key.get_any_name(): output_arr
            for output_key, output_arr in results.items()
        }

        return (results[output_key] for output_key in self.output_keys)


class PytorchOpenVinoInferenceLearner(
    OpenVinoInferenceLearner, PytorchBaseInferenceLearner
):
    """Model optimized using ApacheTVM with a Pytorch interface.

    This class can be used exactly in the same way as a pytorch Module object.
    At prediction time it takes as input pytorch tensors given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        exec_network (any): The graph executor. This is the
            central component in the OpenVino optimized model execution.
        input_keys (List): Keys associated to the inputs.
        output_keys (List): Keys associated to the outputs.
        description_file (str): File containing a description of the optimized
            model.
        weights_file (str): File containing the model weights.
    """

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        input_arrays = (
            input_tensor.cpu().detach().numpy()
            for input_tensor in input_tensors
        )
        output_arrays = self._predict_array(input_arrays)
        return tuple(
            torch.from_numpy(output_array) for output_array in output_arrays
        )


class TensorflowOpenVinoInferenceLearner(
    OpenVinoInferenceLearner, TensorflowBaseInferenceLearner
):
    """Model optimized using ApacheTVM with a tensorflow interface.

    This class can be used exactly in the same way as a tf.Module or
    keras.Model object.
    At prediction time it takes as input tensorflow tensors given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        exec_network (any): The graph executor. This is the
            central component in the OpenVino optimized model execution.
        input_keys (List): Keys associated to the inputs.
        output_keys (List): Keys associated to the outputs.
        description_file (str): File containing a description of the optimized
            model.
        weights_file (str): File containing the model weights.
    """

    def run(self, *input_tensors: tf.Tensor) -> Tuple[tf.Tensor, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        input_arrays = (input_tensor.numpy() for input_tensor in input_tensors)
        output_arrays = self._predict_array(input_arrays)
        # noinspection PyTypeChecker
        return tuple(
            tf.convert_to_tensor(output_array)
            for output_array in output_arrays
        )


class NumpyOpenVinoInferenceLearner(
    OpenVinoInferenceLearner, NumpyBaseInferenceLearner
):
    """Model optimized using ApacheTVM with a numpy interface.

    This class can be used exactly in the same way as a sklearn or
    numpy-based model.
    At prediction time it takes as input numpy arrays given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        exec_network (any): The graph executor. This is the
            central component in the OpenVino optimized model execution.
        input_keys (List): Keys associated to the inputs.
        output_keys (List): Keys associated to the outputs.
        description_file (str): File containing a description of the optimized
            model.
        weights_file (str): File containing the model weights.
    """

    def run(self, *input_tensors: np.ndarray) -> Tuple[np.ndarray, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[np.ndarray]): Input tensors belonging to
                the same batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[np.ndarray]: Output tensors. Note that the output tensors
                does not correspond to the prediction on the input tensors
                with a 1 to 1 mapping. In fact the output tensors are produced
                as the multiple-output of the model given a (multi-) tensor
                input.
        """
        input_arrays = (input_tensor for input_tensor in input_tensors)
        output_arrays = self._predict_array(input_arrays)
        return tuple(output_arrays)


OPENVINO_INFERENCE_LEARNERS: Dict[
    DeepLearningFramework, Type[OpenVinoInferenceLearner]
] = {
    DeepLearningFramework.PYTORCH: PytorchOpenVinoInferenceLearner,
    DeepLearningFramework.TENSORFLOW: TensorflowOpenVinoInferenceLearner,
    DeepLearningFramework.NUMPY: NumpyOpenVinoInferenceLearner,
}



from nebullvm.operations.optimizations.base import Optimizer
from nebullvm.operations.optimizations.compilers.utils import (
    tvm_is_available,
    bladedisc_is_available,
    deepsparse_is_available,
    intel_neural_compressor_is_available,
    torch_tensorrt_is_available,
    onnxruntime_is_available,
    tensorrt_is_available,
    openvino_is_available,
)
from nebullvm.optional_modules.utils import (
    torch_is_available,
    tensorflow_is_available,
    onnx_is_available,
)
from nebullvm.tools.base import DeepLearningFramework, ModelCompiler, Device


class PytorchOptimizer(Optimizer):
    def __init__(self):
        super().__init__()
        self.pipeline_dl_framework = DeepLearningFramework.PYTORCH

    def _select_compilers_from_hardware(self):
        compilers = []
        if torch_is_available():
            compilers.append(ModelCompiler.TORCHSCRIPT)
            if tvm_is_available():
                compilers.append(ModelCompiler.APACHE_TVM)
            if bladedisc_is_available():
                compilers.append(ModelCompiler.BLADEDISC)

            if self.device is Device.CPU:
                if deepsparse_is_available():
                    compilers.append(ModelCompiler.DEEPSPARSE)
                if intel_neural_compressor_is_available():
                    compilers.append(ModelCompiler.INTEL_NEURAL_COMPRESSOR)
            elif self.device is Device.GPU:
                if torch_tensorrt_is_available():
                    compilers.append(ModelCompiler.TENSOR_RT)
        return compilers


class TensorflowOptimizer(Optimizer):
    def __init__(self):
        super().__init__()
        self.pipeline_dl_framework = DeepLearningFramework.TENSORFLOW

    def _select_compilers_from_hardware(self):
        compilers = []
        if tensorflow_is_available():
            compilers.append(ModelCompiler.XLA)
            compilers.append(ModelCompiler.TFLITE)
        return compilers


class ONNXOptimizer(Optimizer):
    def __init__(self):
        super().__init__()
        self.pipeline_dl_framework = DeepLearningFramework.NUMPY

    def _select_compilers_from_hardware(self):
        compilers = []
        if onnx_is_available():
            if onnxruntime_is_available():
                compilers.append(ModelCompiler.ONNX_RUNTIME)
            if tvm_is_available():
                compilers.append(ModelCompiler.APACHE_TVM)
            if self.device is Device.GPU and tensorrt_is_available():
                compilers.append(ModelCompiler.TENSOR_RT)
            if self.device is Device.CPU and openvino_is_available():
                compilers.append(ModelCompiler.OPENVINO)
        return compilers



from typing import List, Tuple

from nebullvm.config import QUANTIZATION_DATA_NUM
from nebullvm.operations.optimizations.compilers.base import Compiler

from nebullvm.operations.optimizations.compilers.quantizations.tensorflow import (  # noqa: E501
    quantize_tensorflow,
)
from nebullvm.operations.optimizations.compilers.quantizations.utils import (
    check_quantization,
)
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.tools.base import QuantizationType
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import MultiStageTransformation


class TensorflowBackendCompiler(Compiler):
    supported_ops = {
        "cpu": [None],
        "gpu": [None],
    }

    def execute(
        self,
        model: tf.Module,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Optimize the input model using tensorflow built-in techniques.

        Args:
            model (tf.Module): The tensorflow model.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction. Default: None.
            metric_drop_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with a higher drop
                will be ignored. Default: None.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used. Default: None.
            input_data (DataManager): User defined data. Default: None.
        """

        if quantization_type not in self.supported_ops[self.device.value]:
            self.compiled_model = None
            return

        if quantization_type is QuantizationType.STATIC and input_data is None:
            raise ValueError("Input data is required for static quantization.")

        self.logger.info(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )

        check_quantization(quantization_type, metric_drop_ths)

        self.compiled_model = model

    def _compile_model(self):
        pass

    @staticmethod
    def _quantize_model(**kwargs):
        raise NotImplementedError()


class TFLiteBackendCompiler(Compiler):
    supported_ops = {
        "cpu": [
            None,
            QuantizationType.STATIC,
            QuantizationType.HALF,
            QuantizationType.DYNAMIC,
        ],
        "gpu": [],
    }

    def execute(
        self,
        model: tf.Module,
        input_tfms: MultiStageTransformation,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Optimize the input model using pytorch built-in techniques.

        Args:
            model (torch.nn.Module): The pytorch model. For avoiding un-wanted
                modifications to the original model, it will be copied in the
                method.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction. Default: None.
            metric_drop_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with an higher drop
                will be ignored. Default: None.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used. Default: None.
            input_data (DataManager): User defined data. Default: None
        """

        if quantization_type not in self.supported_ops[self.device.value]:
            self.compiled_model = None
            return

        if quantization_type is QuantizationType.STATIC and input_data is None:
            raise ValueError("Input data is required for static quantization.")

        self.logger.info(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )

        check_quantization(quantization_type, metric_drop_ths)
        train_input_data = input_data.get_split("train").get_list(
            QUANTIZATION_DATA_NUM
        )

        if quantization_type is not None:
            self.compiled_model = self._quantize_model(
                model, quantization_type, train_input_data
            )
        else:
            self.compiled_model = self._compile_model(model)

    def _compile_model(
        self,
        model: tf.Module,
    ):
        converter = tf.lite.TFLiteConverter.from_keras_model(model)
        tflite_model = converter.convert()
        return tflite_model

    @staticmethod
    def _quantize_model(
        model: tf.Module,
        quantization_type: QuantizationType,
        input_data_tensorflow: List[Tuple[tf.Tensor, ...]],
    ):
        return quantize_tensorflow(
            model, quantization_type, input_data_tensorflow
        )



import copy
from abc import ABC, abstractmethod
from typing import List, Any, Dict

import numpy as np

from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch


class BaseTransformation(ABC):
    @abstractmethod
    def _transform(self, _input: Any, **kwargs) -> Any:
        raise NotImplementedError()

    def __call__(self, _input: Any, **kwargs):
        return self._transform(_input, **kwargs)

    def to_dict(self):
        return {
            "module": self.__class__.__module__,
            "name": self.__class__.__name__,
        }

    @classmethod
    def from_dict(cls, tfm_dict: Dict):
        return cls()


class MultiStageTransformation(BaseTransformation):
    def __init__(self, transformations: List[BaseTransformation]):
        self._tfms = transformations

    def _transform(self, _input: Any, **kwargs) -> Any:
        for tfm in self._tfms:
            _input = tfm(_input, **kwargs)
        return _input

    def append(self, __tfm: BaseTransformation):
        self._tfms.append(__tfm)

    def extend(self, tfms: List[BaseTransformation]):
        self._tfms += tfms

    def to_dict(self) -> Dict:
        return {"tfms": [tfm.to_dict() for tfm in self._tfms]}

    def to_list(self):
        return self._tfms

    @classmethod
    def from_dict(cls, tfms_dict: Dict):
        tfms = []
        for tfm_dict in tfms_dict["tfms"]:
            exec(f"from {tfm_dict['module']} import {tfm_dict['name']}")
            tfm = eval(tfm_dict["name"]).from_dict(tfm_dict)
            tfms.append(tfm)
        return cls(tfms)

    def copy(self):
        new_list = copy.deepcopy(self._tfms)
        return self.__class__(new_list)

    def __len__(self):
        return len(self._tfms)


class HalfPrecisionTransformation(BaseTransformation):
    @staticmethod
    def _transform_numpy(_input: np.ndarray) -> np.ndarray:
        return _input.astype(dtype=np.float16)

    @staticmethod
    def _transform_tf(_input: tf.Tensor) -> tf.Tensor:
        return tf.cast(_input, tf.float16)

    @staticmethod
    def _transform_torch(_input: torch.Tensor) -> torch.Tensor:
        return _input.half()

    def _transform(self, _input: Any, **kwargs) -> Any:
        if isinstance(_input, np.ndarray):
            return (
                self._transform_numpy(_input)
                if _input.dtype == np.float32
                else _input
            )
        elif isinstance(_input, torch.Tensor):
            return (
                self._transform_torch(_input)
                if _input.dtype == torch.float32
                else _input
            )
        elif isinstance(_input, tf.Tensor) and _input is not None:
            return (
                self._transform_tf(_input)
                if _input.dtype == tf.float32
                else _input
            )
        else:
            raise TypeError(
                f"The given input type is not currently supported. "
                f"Got {type(_input)}, expected one between (np.ndarray, "
                f"torch.Tensor, tf.Tensor)"
            )


class NoOp(BaseTransformation):
    def _transform(self, _input: Any, **kwargs):
        return _input


class VerifyContiguity(BaseTransformation):
    def _transform(self, _input: Any, **kwargs) -> Any:
        if not isinstance(_input, torch.Tensor):
            return _input
        if not _input.is_contiguous():
            _input = _input.contiguous()
        return _input



try:
    import neural_compressor  # noqa F401
    from neural_compressor.adaptor.pytorch import (
        _cfg_to_qconfig as cfg_to_qconfig,
        _cfgs_to_fx_cfgs as cfgs_to_fx_cfgs,
    )
    from neural_compressor.experimental import (
        MixedPrecision,
        Quantization,
        Pruning,
    )
except ImportError:
    cfg_to_qconfig = cfgs_to_fx_cfgs = None
    MixedPrecision = Quantization = Pruning = object
except ValueError:
    # MacOS
    MixedPrecision = Quantization = Pruning = object



import copy
from typing import List, Tuple, Union

from loguru import logger

from nebullvm.optional_modules.torch import (
    torch,
    Module,
    symbolic_trace,
    QuantStub,
    DeQuantStub,
    GraphModule,
    default_dynamic_qconfig,
    prepare_fx,
    convert_fx,
    ScriptModule,
)
from nebullvm.tools.base import Device, QuantizationType
from nebullvm.tools.transformations import (
    MultiStageTransformation,
    HalfPrecisionTransformation,
)
from nebullvm.tools.utils import check_module_version


class _QuantWrapper(Module):
    def __init__(self, model: Module):
        super(_QuantWrapper, self).__init__()
        qconfig = model.qconfig if hasattr(model, "qconfig") else None
        self.quant = QuantStub(qconfig)
        self.model = model
        self.dequant = DeQuantStub()

    def forward(self, *inputs: torch.Tensor):
        inputs = (self.quant(x) for x in inputs)
        outputs = self.model(*inputs)
        return tuple(self.dequant(x) for x in outputs)


def _quantize_dynamic_torch(model: Module):
    layer_types = {
        type(layer)
        for layer in model.children()
        if len(list(layer.parameters())) > 0
    }
    return torch.quantization.quantize_dynamic(
        model=model, qconfig_spec=layer_types, dtype=torch.qint8
    )


def _quantize_dynamic_torch_fx(
    model: GraphModule,
    input_data: List[Tuple[torch.Tensor, ...]],
):
    qconfig_dict = {"": default_dynamic_qconfig}

    additional_arguments = {}
    if check_module_version(torch, min_version="1.13.0"):
        additional_arguments["example_inputs"] = input_data[0]

    model_prepared = prepare_fx(model, qconfig_dict, **additional_arguments)
    return convert_fx(model_prepared)


def _quantize_static_torch(
    model: Module,
    input_data: List[Tuple[torch.Tensor, ...]],
    backend: str,
):
    model = _QuantWrapper(model)
    model.qconfig = torch.quantization.get_default_qconfig(backend)
    # TODO: change line below, it's wrong
    # model = torch.quantization.fuse_modules(model, [["conv", "relu"]])
    model = torch.quantization.prepare(model)
    with torch.no_grad():
        for tensors in input_data:
            _ = model(*tensors)
    return torch.quantization.convert(model)


def _quantize_static_torch_fx(
    model: GraphModule,
    input_data: List[Tuple[torch.Tensor, ...]],
    backend: str,
):
    qconfig_dict = {"": torch.quantization.get_default_qconfig(backend)}
    additional_arguments = {}
    if check_module_version(torch, min_version="1.13.0"):
        additional_arguments["example_inputs"] = input_data[0]

    model_prepared = prepare_fx(model, qconfig_dict, **additional_arguments)
    with torch.no_grad():
        for tensors in input_data:
            _ = model_prepared(*tensors)
    return convert_fx(model_prepared)


def _quantize_static(
    model: Union[Module, GraphModule],
    input_data: List[Tuple[torch.Tensor, ...]],
    device: Device,
):
    assert (
        device is not Device.GPU
    ), "Quantization for torch is only available on CPU"

    backend = (
        "fbgemm"
        if "fbgemm" in torch.backends.quantized.supported_engines
        else "qnnpack"
    )

    torch.backends.quantized.engine = backend

    if isinstance(model, GraphModule):
        return _quantize_static_torch_fx(model, input_data, backend)
    else:
        return _quantize_static_torch(model, input_data, backend)


def _quantize_dynamic(
    model: Union[Module, GraphModule],
    input_data: List[Tuple[torch.Tensor, ...]],
    device: Device,
):
    assert (
        device is not Device.GPU
    ), "Quantization for torch is only available on CPU"

    backend = (
        "fbgemm"
        if "fbgemm" in torch.backends.quantized.supported_engines
        else "qnnpack"
    )

    torch.backends.quantized.engine = backend

    if isinstance(model, GraphModule):
        return _quantize_dynamic_torch_fx(model, input_data)
    else:
        return _quantize_dynamic_torch(model)


def _half_precision(model: Module):
    return model.half()


def quantize_pytorch(
    model: Module,
    quantization_type: QuantizationType,
    input_tfms: MultiStageTransformation,
    input_data_torch: List[Tuple[torch.Tensor, ...]],
    device: Device,
) -> Union[torch.nn.Module, ScriptModule, GraphModule]:
    model = copy.deepcopy(model).eval()

    try:
        model = symbolic_trace(model)
    except Exception:
        logger.warning("Unable to trace model with torch.fx")

    if quantization_type is QuantizationType.HALF:
        input_tfms.append(HalfPrecisionTransformation())
        quantized_model = _half_precision(model)
    elif quantization_type is QuantizationType.STATIC:
        quantized_model = _quantize_static(model, input_data_torch, device)
    elif quantization_type is QuantizationType.DYNAMIC:
        quantized_model = _quantize_dynamic(model, input_data_torch, device)
    else:
        raise NotImplementedError(
            f"No quantization implemented for quantization "
            f"type {quantization_type}"
        )

    return quantized_model



try:
    import torch  # noqa F401
    from torch.nn import Module  # noqa F401
    from torch.jit import ScriptModule  # noqa F401
    from torch.fx import GraphModule
    from torch.utils.data import DataLoader, Dataset  # noqa F401
    from torch.quantization.quantize_fx import (  # noqa F401
        prepare_fx,
        convert_fx,
    )

    from torch.ao.quantization.stubs import QuantStub, DeQuantStub
    from torch.fx import symbolic_trace
    from torch.quantization import default_dynamic_qconfig
except ImportError:

    class Tensor:
        pass

    class Module:
        pass

    class nn:
        Module = Module

    class torch:
        float = half = int8 = None
        Tensor = Tensor
        dtype = None
        nn = nn

        @staticmethod
        def no_grad():
            return lambda x: None

    class ScriptModule:
        pass

    class GraphModule:
        pass

    Dataset = object
    DataLoader = type(None)
    symbolic_trace = None
    QuantStub = (
        DeQuantStub
    ) = default_dynamic_qconfig = prepare_fx = convert_fx = None



from typing import List, Tuple, Any, Callable, Dict, Optional

import numpy as np

from nebullvm.config import QUANTIZATION_DATA_NUM
from nebullvm.operations.inference_learners.base import BaseInferenceLearner
from nebullvm.operations.measures.base import Measure
from nebullvm.operations.measures.utils import (
    compute_torch_latency,
    compute_tf_latency,
    compute_onnx_latency,
    compute_relative_difference,
)
from nebullvm.tools.base import DeepLearningFramework
from nebullvm.tools.data import DataManager
from nebullvm.tools.onnx import run_onnx_model
from nebullvm.tools.pytorch import run_torch_model
from nebullvm.tools.tf import run_tf_model

COMPUTE_OUTPUT_FRAMEWORK: Dict[DeepLearningFramework, Callable] = {
    DeepLearningFramework.PYTORCH: run_torch_model,
    DeepLearningFramework.TENSORFLOW: run_tf_model,
    DeepLearningFramework.NUMPY: run_onnx_model,
}

COMPUTE_LATENCY_FRAMEWORK: Dict[DeepLearningFramework, Callable] = {
    DeepLearningFramework.PYTORCH: compute_torch_latency,
    DeepLearningFramework.TENSORFLOW: compute_tf_latency,
    DeepLearningFramework.NUMPY: compute_onnx_latency,
}


class MetricDropMeasure(Measure):
    def __init__(self):
        super().__init__()
        self.valid = None

    def execute(
        self,
        optimized_learner: BaseInferenceLearner,
        input_data: List[Tuple[Any, ...]],
        base_outputs_list: List[Tuple[Any, ...]],
        perf_loss_ths: float,
        metric_func: Callable = None,
        ys: List = None,
        aggregation_func: Callable = np.mean,
    ):
        metric_func = metric_func or compute_relative_difference
        relative_differences = []
        if ys is None:
            ys = [None] * len(input_data)

        assert len(input_data) == len(base_outputs_list) == len(ys), (
            "INTERNAL ASSERT FAILED: error during computation of precision "
            "of the optimized model, got wrong dimensions of the data. "
        )

        for inputs, base_outputs, y in zip(input_data, base_outputs_list, ys):
            opt_outputs = optimized_learner(*inputs)
            relative_difference = max(
                metric_func(base_output, opt_output, y)
                for base_output, opt_output in zip(base_outputs, opt_outputs)
            )
            relative_differences.append(relative_difference)
        relative_difference = aggregation_func(relative_differences)
        self.valid = relative_difference <= perf_loss_ths
        self.measure_result = relative_difference

    def get_result(self) -> Tuple:
        return self.valid, self.measure_result


class LatencyOriginalModelMeasure(Measure):
    def __init__(self):
        super().__init__()
        self.outputs = None

    def execute(
        self,
        model: Any,
        input_data: DataManager,
        dl_framework: DeepLearningFramework,
    ):
        self.logger.info("Benchmark performance of original model")

        self.outputs = [
            tuple(
                COMPUTE_OUTPUT_FRAMEWORK[dl_framework](
                    model, tuple(input_tensors[0]), self.device
                )
            )
            for input_tensors in input_data
        ]

        inputs = input_data.get_list(QUANTIZATION_DATA_NUM)
        self.measure_result, _ = COMPUTE_LATENCY_FRAMEWORK[dl_framework](
            inputs, model, self.device
        )
        self.logger.info(
            f"Original model latency: {self.measure_result} sec/iter"
        )

    def get_result(self) -> Optional[Tuple]:
        if self.outputs is not None and self.measure_result is not None:
            return self.outputs, self.measure_result
        else:
            return None



import abc
from typing import Dict, Any, Optional, Union

from loguru import logger

from nebullvm.tools.base import Device
from nebullvm.tools.feedback_collector import FeedbackCollector
from nebullvm.tools.utils import gpu_is_available


def _check_device(device: Optional[str]) -> Device:
    if device is None:
        if gpu_is_available():
            device = Device.GPU
        else:
            device = Device.CPU
    else:
        if device.lower() == "gpu":
            if not gpu_is_available():
                logger.warning(
                    "Selected GPU device but no available GPU found on this "
                    "platform. CPU will be used instead. Please make sure "
                    "that the gpu is installed and can be used by your "
                    "framework."
                )
                device = Device.CPU
            else:
                device = Device.GPU
        else:
            device = Device.CPU

    return device


class Operation(abc.ABC):
    def __init__(self):
        self._state = {}
        self.device = Device.CPU
        self.execute_count = 0
        self.logger = logger
        self.feedback_collector = None

    def set_feedback_collector(self, feedback_collector: FeedbackCollector):
        self.feedback_collector = feedback_collector
        for value in self.__dict__.values():
            if isinstance(value, Operation):
                value.set_feedback_collector(feedback_collector)

    @abc.abstractmethod
    def execute(self, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_result(self) -> Any:
        raise NotImplementedError()

    @property
    def state(self) -> Dict[str, any]:
        return self._state

    def to(self, device: Union[str, Device]):
        if isinstance(device, str):
            self.device = _check_device(device)
        else:
            self.device = device
        return self



import json
import os
from abc import ABC
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Union, Dict, Type, List, Tuple, Generator, Optional

import numpy as np
from loguru import logger

from nebullvm.config import NVIDIA_FILENAMES
from nebullvm.operations.inference_learners.base import (
    BaseInferenceLearner,
    LearnerMetadata,
    PytorchBaseInferenceLearner,
    TensorflowBaseInferenceLearner,
    NumpyBaseInferenceLearner,
)
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.tensor_rt import tensorrt as trt, polygraphy
from nebullvm.optional_modules.torch import torch, ScriptModule
from nebullvm.tools.base import Device, ModelParams, DeepLearningFramework
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import (
    MultiStageTransformation,
    VerifyContiguity,
)


class ONNXTensorRTInferenceLearner(BaseInferenceLearner, ABC):
    """Model optimized using TensorRT.

    The class cannot be directly instantiated, but implements all the core
    methods needed for using TensorRT at inference time.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        engine (any): The tensorRT engine.
        input_names (List[str]): Names associated to the model input tensors.
        output_names (List[str]): Names associated to the model output tensors.
        cuda_stream (any, optional): Stream used for communication with Nvidia
            GPUs.
        nvidia_logger (any, optional): Logger used by the Nvidia service
    """

    name = "TensorRT"

    def __init__(
        self,
        engine: Any,
        input_names: List[str],
        output_names: List[str],
        device: Device,
        cuda_stream: Any = None,
        nvidia_logger: Any = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.engine = engine
        self.input_names = input_names
        self.output_names = output_names
        self.cuda_stream = cuda_stream
        self.nvidia_logger = nvidia_logger
        self._set_cuda_env(device is Device.GPU)

    def _get_metadata(self, **kwargs) -> LearnerMetadata:
        metadata = {
            key: self.__dict__[key] for key in ("input_names", "output_names")
        }
        metadata.update(kwargs)
        return LearnerMetadata.from_model(self, **metadata)

    def _synchronize_stream(self):
        raise NotImplementedError()

    @property
    def stream_ptr(self):
        raise NotImplementedError()

    @staticmethod
    def _get_default_cuda_stream() -> Any:
        raise NotImplementedError()

    @staticmethod
    def check_env(use_gpu):
        if not use_gpu:
            raise SystemError(
                "You are trying to run an optimizer developed for NVidia gpus "
                "on a machine not connected to any GPU supporting CUDA."
            )

    def _set_cuda_env(self, use_gpu):
        self.check_env(use_gpu)
        if self.nvidia_logger is None:
            self.nvidia_logger = trt.Logger(trt.Logger.WARNING)
        if self.cuda_stream is None:
            self.cuda_stream = self._get_default_cuda_stream()

    @classmethod
    def from_engine_path(
        cls,
        network_parameters: ModelParams,
        engine_path: Union[str, Path],
        input_names: List[str],
        output_names: List[str],
        device: Device,
        nvidia_logger: Any = None,
        cuda_stream: Any = None,
        input_tfms: MultiStageTransformation = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Build the model from the serialised engine.

        Args:
            network_parameters (ModelParams): Model parameters.
            engine_path (str or Path): Path to the serialised engine. The
                serialised engine is the serialised version of the engine
                used for accelerating the inference.
            input_names (List[str]): Names associated to the model input
                tensors.
            output_names (List[str]): Names associated to the model output
                tensors.
            device: (Device): Device where the model wil be run.
            cuda_stream (any, optional): Stream used for communication with
                Nvidia GPUs.
            nvidia_logger (any, optional): Logger used by the Nvidia service
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            input_data (DataManager, optional): User defined data.

        Returns:
            NvidiaInferenceLearner: The optimized model.
        """
        if kwargs:
            logger.warning(
                f"Debug: Got extra keywords in "
                f"NvidiaInferenceLearner::from_engine_path: {kwargs}"
            )
        if nvidia_logger is None:
            nvidia_logger = trt.Logger(trt.Logger.WARNING)
        if input_tfms is None:
            input_tfms = MultiStageTransformation([])
        input_tfms.append(VerifyContiguity())
        runtime = trt.Runtime(nvidia_logger)
        with open(engine_path, "rb") as f:
            serialized_engine = f.read()
        engine = runtime.deserialize_cuda_engine(serialized_engine)
        return cls(
            input_tfms=input_tfms,
            network_parameters=network_parameters,
            engine=engine,
            input_names=input_names,
            output_names=output_names,
            nvidia_logger=nvidia_logger,
            cuda_stream=cuda_stream,
            input_data=input_data,
            device=device,
        )

    def _predict_tensors(
        self,
        input_ptrs: Generator[Any, None, None],
        output_ptrs: Generator[Any, None, None],
        input_shapes: Generator[Any, None, None] = None,
    ):
        context = self.engine.create_execution_context()
        buffers = [None] * (len(self.input_names) + len(self.output_names))
        input_idxs = (
            self.engine[input_name] for input_name in self.input_names
        )
        output_idxs = (
            self.engine[output_name] for output_name in self.output_names
        )
        input_shapes = input_shapes or [None] * len(self.input_names)
        for input_idx, input_ptr, input_shape in zip(
            input_idxs, input_ptrs, input_shapes
        ):
            buffers[input_idx] = input_ptr
            if input_shape is not None:
                context.set_binding_shape(input_idx, input_shape)
        for output_idx, output_ptr in zip(output_idxs, output_ptrs):
            buffers[output_idx] = output_ptr
        context.execute_async_v2(buffers, self.stream_ptr)
        self._synchronize_stream()

    def get_size(self):
        return self.engine.serialize().nbytes

    def save(self, path: Union[str, Path], **kwargs):
        """Save the model.

        Args:
            path (Path or str): Path to the directory where the model will
                be stored.
            kwargs (Dict): Dictionary of key-value pairs that will be saved in
                the model metadata file.
        """
        path = Path(path)
        path.mkdir(exist_ok=True)
        serialized_engine = self.engine.serialize()
        with open(path / NVIDIA_FILENAMES["engine"], "wb") as fout:
            fout.write(serialized_engine)
        metadata = self._get_metadata(**kwargs)
        with open(path / NVIDIA_FILENAMES["metadata"], "w") as fout:
            json.dump(metadata.to_dict(), fout)

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        """Load the model.

        Args:
            path (Path or str): Path to the directory where the model is
                stored.
            kwargs (Dict): Dictionary of additional arguments for the
                `from_engine_path` class method.

        Returns:
            ONNXTensorRTInferenceLearner: The optimized model.
        """
        path = Path(path)
        with open(path / NVIDIA_FILENAMES["metadata"], "r") as fin:
            metadata = json.load(fin)
        metadata.update(kwargs)
        metadata["network_parameters"] = ModelParams(
            **metadata["network_parameters"]
        )
        input_tfms = metadata.get("input_tfms")
        if input_tfms is not None:
            metadata["input_tfms"] = MultiStageTransformation.from_dict(
                input_tfms
            )
        device = Device.GPU
        return cls.from_engine_path(
            engine_path=path / NVIDIA_FILENAMES["engine"],
            device=device,
            **metadata,
        )


class PytorchTensorRTInferenceLearner(PytorchBaseInferenceLearner):
    MODEL_NAME = "model_optimized.pt"
    name = "TensorRT"

    def __init__(
        self,
        torch_model: ScriptModule,
        device: Device,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.model = torch_model.eval()
        if device is Device.GPU:
            self.model.cuda()
            self.use_gpu = True
        else:
            self.use_gpu = False

    def get_size(self):
        with TemporaryDirectory() as tmp_dir:
            self.save(tmp_dir)
            return sum(
                os.path.getsize(Path(tmp_dir) / f)
                for f in os.listdir(Path(tmp_dir))
                if os.path.isfile(Path(tmp_dir) / f)
            )

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        device = input_tensors[0].device

        # PyTorch-TensorRT does not support int64
        input_tensors = (
            t.cuda() if t.dtype != torch.int64 else t.to(torch.int32).cuda()
            for t in input_tensors
        )

        with torch.no_grad():
            res = self.model(*input_tensors)
            if not isinstance(res, tuple):
                res = res.to(device)
                return (res,)
            return tuple(out.to(device) for out in res)

    def save(self, path: Union[str, Path], **kwargs):
        path = Path(path)
        path.mkdir(exist_ok=True)
        metadata = LearnerMetadata.from_model(self, **kwargs)
        metadata.save(path)
        torch.jit.save(self.model, path / self.MODEL_NAME)

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        path = Path(path)
        model = torch.jit.load(path / cls.MODEL_NAME)
        metadata = LearnerMetadata.read(path)
        device = Device.GPU
        return cls(
            torch_model=model,
            network_parameters=ModelParams(**metadata.network_parameters),
            input_tfms=MultiStageTransformation.from_dict(metadata.input_tfms)
            if metadata.input_tfms is not None
            else None,
            device=device,
        )


class PytorchONNXTensorRTInferenceLearner(
    ONNXTensorRTInferenceLearner, PytorchBaseInferenceLearner
):
    """Model optimized using TensorRT with a Pytorch interface.

    This class can be used exactly in the same way as a pytorch Module object.
    At prediction time it takes as input pytorch tensors given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        engine (any): The tensorRT engine.
        input_names (List[str]): Names associated to the model input tensors.
        output_names (List[str]): Names associated to the model output tensors.
        cuda_stream (any, optional): Stream used for communication with Nvidia
            GPUs.
        nvidia_logger (any, optional): Logger used by the Nvidia service.
    """

    def _synchronize_stream(self):
        self.cuda_stream.synchronize()

    @staticmethod
    def _get_default_cuda_stream() -> Any:
        return torch.cuda.default_stream()

    @property
    def stream_ptr(self):
        return self.cuda_stream.cuda_stream

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        input_tensors = [input_tensor.cuda() for input_tensor in input_tensors]
        device = input_tensors[0].device
        if self.network_parameters.dynamic_info is None:
            output_tensors = [
                torch.Tensor(
                    self.network_parameters.batch_size,
                    *output_size,
                ).cuda()
                for output_size in self.network_parameters.output_sizes
            ]
            input_sizes = None
        else:
            dynamic_info = self.network_parameters.dynamic_info
            input_sizes = [
                input_tensor.size() for input_tensor in input_tensors
            ]
            output_tensors = [
                torch.Tensor(
                    *(
                        x
                        if i not in dynamic_axis.keys()
                        else dynamic_info.retrieve_output_dim(
                            input_sizes, j, i, x
                        )
                        for i, x in enumerate(
                            (self.network_parameters.batch_size,) + output_size
                        )
                    ),
                ).cuda()
                for j, (output_size, dynamic_axis) in enumerate(
                    zip(
                        self.network_parameters.output_sizes,
                        dynamic_info.outputs,
                    )
                )
            ]
        input_ptrs = (
            input_tensor.data_ptr() for input_tensor in input_tensors
        )
        output_ptrs = (
            output_tensor.data_ptr() for output_tensor in output_tensors
        )
        self._predict_tensors(input_ptrs, output_ptrs, input_sizes)
        return tuple(
            output_tensor.to(device) for output_tensor in output_tensors
        )


class BaseArrayONNXTensorRTInferenceLearner(ONNXTensorRTInferenceLearner, ABC):
    """Base Model that can be used for all array-based
    NvidiaInferenceLearners.
    """

    def _synchronize_stream(self):
        self.cuda_stream.synchronize()

    @staticmethod
    def _get_default_cuda_stream() -> Any:
        return polygraphy.cuda.Stream()

    @property
    def stream_ptr(self):
        return self.cuda_stream.ptr

    @staticmethod
    def _convert_to_array_and_free_memory(cuda_array) -> np.ndarray:
        array = cuda_array.numpy()
        cuda_array.free()
        return array

    def _predict_array(
        self,
        cuda_input_arrays: List,
        input_shapes: Optional[List[Tuple[int, ...]]],
    ) -> Generator[np.ndarray, None, None]:
        if self.network_parameters.dynamic_info is None:
            cuda_output_arrays = [
                polygraphy.cuda.DeviceArray(
                    shape=(self.network_parameters.batch_size, *output_size)
                )
                for output_size in self.network_parameters.output_sizes
            ]
        else:
            dynamic_info = self.network_parameters.dynamic_info
            output_sizes = (
                (self.network_parameters.batch_size, *output_size)
                for output_size in self.network_parameters.output_sizes
            )

            cuda_output_arrays = [
                polygraphy.cuda.DeviceArray(
                    shape=tuple(
                        x
                        if i not in dyn_out_axis.keys()
                        else dynamic_info.retrieve_output_dim(
                            input_shapes, j, i, x
                        )
                        for i, x in enumerate(output_size)
                    )
                )
                for j, (output_size, dyn_out_axis) in enumerate(
                    zip(output_sizes, dynamic_info.outputs)
                )
            ]
        input_ptrs = (cuda_array.ptr for cuda_array in cuda_input_arrays)
        output_ptrs = (cuda_array.ptr for cuda_array in cuda_output_arrays)
        self._predict_tensors(input_ptrs, output_ptrs, input_shapes)
        for cuda_input_array in cuda_input_arrays:
            cuda_input_array.free()
        return (
            self._convert_to_array_and_free_memory(array)
            for array in cuda_output_arrays
        )


class TensorflowONNXTensorRTInferenceLearner(
    BaseArrayONNXTensorRTInferenceLearner, TensorflowBaseInferenceLearner
):
    """Model optimized using TensorRT with a tensorflow interface.

    This class can be used exactly in the same way as a tf.Module or
    keras.Model object.
    At prediction time it takes as input tensorflow tensors given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        engine (any): The tensorRT engine.
        input_names (List[str]): Names associated to the model input tensors.
        output_names (List[str]): Names associated to the model output tensors.
        cuda_stream (any, optional): Stream used for communication with Nvidia
            GPUs.
        nvidia_logger (any, optional): Logger used by the Nvidia service.
    """

    def run(self, *input_tensors: tf.Tensor) -> Tuple[tf.Tensor, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        cuda_input_arrays = [
            polygraphy.cuda.DeviceArray(
                shape=tuple(input_tensor.shape),
                dtype=input_tensor.numpy().dtype,
            ).copy_from(input_tensor.numpy(), stream=self.cuda_stream)
            for input_tensor in input_tensors
        ]
        input_shapes = (
            [tuple(input_tensor.shape) for input_tensor in input_tensors]
            if self.network_parameters.dynamic_info is not None
            else None
        )
        out_arrays = self._predict_array(cuda_input_arrays, input_shapes)
        return tuple(tf.convert_to_tensor(array[0]) for array in out_arrays)


class NumpyONNXTensorRTInferenceLearner(
    BaseArrayONNXTensorRTInferenceLearner, NumpyBaseInferenceLearner
):
    """Model optimized using TensorRT with a tensorflow interface.

    This class can be used exactly in the same way as a tf.Module or
    keras.Model object.
    At prediction time it takes as input tensorflow tensors given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        engine (any): The tensorRT engine.
        input_names (List[str]): Names associated to the model input tensors.
        output_names (List[str]): Names associated to the model output tensors.
        cuda_stream (any, optional): Stream used for communication with Nvidia
            GPUs.
        nvidia_logger (any, optional): Logger used by the Nvidia service.
    """

    def run(self, *input_tensors: np.ndarray) -> Tuple[np.ndarray, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[np.ndarray]): Input tensors belonging to
                the same batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[np.ndarray]: Output tensors. Note that the output tensors
                does not correspond to the prediction on the input tensors
                with a 1 to 1 mapping. In fact the output tensors are produced
                as the multiple-output of the model given a (multi-) tensor
                input.
        """
        cuda_input_arrays = [
            polygraphy.cuda.DeviceArray(
                shape=tuple(input_tensor.shape), dtype=input_tensor.dtype
            ).copy_from(input_tensor, stream=self.cuda_stream)
            for input_tensor in input_tensors
        ]
        input_shapes = (
            [tuple(input_tensor.shape) for input_tensor in input_tensors]
            if self.network_parameters.dynamic_info is not None
            else None
        )
        return tuple(self._predict_array(cuda_input_arrays, input_shapes))


TENSOR_RT_INFERENCE_LEARNERS: Dict[
    DeepLearningFramework, Type[ONNXTensorRTInferenceLearner]
] = {
    DeepLearningFramework.PYTORCH: PytorchONNXTensorRTInferenceLearner,
    DeepLearningFramework.TENSORFLOW: TensorflowONNXTensorRTInferenceLearner,
    DeepLearningFramework.NUMPY: NumpyONNXTensorRTInferenceLearner,
}



from typing import List, Callable


def map_compilers_and_compressors(ignore_list: List, enum_class: Callable):
    if ignore_list is None:
        ignore_list = []
    else:
        ignore_list = [enum_class(element) for element in ignore_list]
    return ignore_list



from nebullvm.installers.auto_installer import check_backends


def test_install_all_backends():
    include_backends = "all"

    include_backends = check_backends(include_backends)

    assert include_backends == ["huggingface", "onnx", "tensorflow", "torch"]


def test_install_torch_full():
    include_backends = ["torch-full"]

    include_backends = check_backends(include_backends)

    assert include_backends == ["onnx", "torch"]


def test_install_torch_base():
    include_backends = ["torch-base"]

    include_backends = check_backends(include_backends)

    assert include_backends == ["torch"]


def test_install_tensorflow_full():
    include_backends = ["tensorflow-full"]

    include_backends = check_backends(include_backends)

    assert include_backends == ["onnx", "tensorflow"]


def test_install_tensorflow_base():
    include_backends = ["tensorflow-base"]

    include_backends = check_backends(include_backends)

    assert include_backends == ["tensorflow"]


def test_install_onnx_full():
    include_backends = ["onnx-full"]

    include_backends = check_backends(include_backends)

    assert include_backends == ["onnx"]


def test_install_onnx_base():
    include_backends = ["onnx-base"]

    include_backends = check_backends(include_backends)

    assert include_backends == ["onnx"]


def test_install_huggingface_full():
    include_backends = ["huggingface-full"]

    include_backends = check_backends(include_backends)

    assert include_backends == ["huggingface", "onnx", "tensorflow", "torch"]


def test_install_huggingface_full_tf():
    include_backends = ["huggingface-full-tf"]

    include_backends = check_backends(include_backends)

    assert include_backends == ["huggingface", "onnx", "tensorflow"]


def test_install_huggingface_full_torch():
    include_backends = ["huggingface-full-torch"]

    include_backends = check_backends(include_backends)

    assert include_backends == ["huggingface", "onnx", "torch"]


def test_install_huggingface_tf():
    include_backends = ["huggingface-base-tf"]

    include_backends = check_backends(include_backends)

    assert include_backends == ["huggingface", "tensorflow"]


def test_install_huggingface_torch():
    include_backends = ["huggingface-base-torch"]

    include_backends = check_backends(include_backends)

    assert include_backends == ["huggingface", "torch"]



from typing import Optional

from nebullvm.operations.inference_learners.pytorch import (
    PytorchBackendInferenceLearner,
)
from nebullvm.optional_modules.torch import ScriptModule
from nebullvm.tools.base import ModelParams, Device
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import MultiStageTransformation


class BladeDISCInferenceLearner(PytorchBackendInferenceLearner):
    name = "BladeDISC"

    @classmethod
    def from_torch_model(
        cls,
        model: ScriptModule,
        network_parameters: ModelParams,
        device: Device,
        input_tfms: Optional[MultiStageTransformation] = None,
        input_data: DataManager = None,
    ):
        return cls(
            torch_model=model,
            network_parameters=network_parameters,
            input_tfms=input_tfms,
            input_data=input_data,
            device=device,
        )



import argparse
from typing import List, Union

from loguru import logger

from nebullvm.config import (
    ONNX_MODULES,
    TENSORFLOW_MODULES,
    TORCH_MODULES,
    HUGGING_FACE_MODULES,
)
from nebullvm.installers.installers import (
    ONNXInstaller,
    PytorchInstaller,
    TensorflowInstaller,
    HuggingFaceInstaller,
)


SUPPORTED_BACKENDS = [
    "torch-full",
    "torch-base",
    "tensorflow-full",
    "tensorflow-base",
    "onnx-full",
    "onnx-base",
    "huggingface-full",
    "huggingface-full-tf",
    "huggingface-full-torch",
    "huggingface-base-tf",
    "huggingface-base-torch",
]

INSTALLERS = {
    "onnx": ONNXInstaller,
    "torch": PytorchInstaller,
    "tensorflow": TensorflowInstaller,
    "huggingface": HuggingFaceInstaller,
}

MODULES = {
    "onnx": ONNX_MODULES,
    "torch": TORCH_MODULES,
    "tensorflow": TENSORFLOW_MODULES,
    "huggingface": HUGGING_FACE_MODULES,
}


def check_backends(include_backends: Union[List[str], str]) -> List[str]:
    if isinstance(include_backends, str) and include_backends == "all":
        new_include_backends = list(INSTALLERS.keys())
    else:
        new_include_backends = []
        for backend in include_backends:
            if backend not in SUPPORTED_BACKENDS:
                raise ValueError(
                    f"Backend {backend} is not supported by nebullvm. "
                    f"Please check the docs to see all the supported options."
                )

            elif backend.startswith("torch"):
                new_include_backends.append("torch")
                if backend.endswith("full"):
                    new_include_backends.append("onnx")
            elif backend.startswith("tensorflow"):
                new_include_backends.append("tensorflow")
                if backend.endswith("full"):
                    new_include_backends.append("onnx")
            elif backend.startswith("huggingface"):
                new_include_backends.append("huggingface")
                if backend.endswith("full") or backend.endswith("tf"):
                    new_include_backends.append("tensorflow")
                if backend.endswith("full") or backend.endswith("torch"):
                    new_include_backends.append("torch")
                if "full" in backend:
                    new_include_backends.append("onnx")
            elif backend.startswith("onnx"):
                new_include_backends.append("onnx")

    # Remove duplicates
    new_include_backends = list(set(new_include_backends))
    new_include_backends.sort()

    return new_include_backends


def auto_install_libraries(
    include_backends: Union[List[str], str] = "all",
    include_compilers: Union[List[str], str] = "all",
):
    logger.info("Running auto install of nebullvm dependencies")

    include_backends = check_backends(include_backends)

    for backend in include_backends:
        backend_installer = INSTALLERS[backend](MODULES[backend])
        if not backend_installer.check_backend():
            backend_installer.install_backend()
        backend_installer.install_dependencies(include_backends)
        backend_installer.install_compilers(include_compilers)


def main():
    parser = argparse.ArgumentParser(
        description="Auto install dl backends and dependencies"
    )
    parser.add_argument(
        "-f",
        "--backends",
        help="The dl backends whose compilers will be installed",
        default="all",
        nargs="+",
    )
    parser.add_argument(
        "-c",
        "--compilers",
        help="Compilers to be installed",
        default="all",
        nargs="+",
    )
    args = vars(parser.parse_args())

    if len(args["backends"]) == 1 and args["backends"][0] == "all":
        backend_list = "all"
    else:
        backend_list = args["backends"]

    if len(args["compilers"]) == 1 and args["compilers"][0] == "all":
        compilers_list = "all"
    else:
        compilers_list = args["compilers"]

    auto_install_libraries(backend_list, compilers_list)


if __name__ == "__main__":
    main()



from tempfile import TemporaryDirectory

import pytest

from nebullvm.operations.inference_learners.tensorflow import (
    TensorflowBackendInferenceLearner,
    TFLiteBackendInferenceLearner,
)
from nebullvm.operations.optimizations.base import (
    COMPILER_TO_INFERENCE_LEARNER_MAP,
)
from nebullvm.operations.optimizations.compilers.tensorflow import (
    TensorflowBackendCompiler,
    TFLiteBackendCompiler,
)
from nebullvm.operations.optimizations.tests.utils import (
    initialize_model,
    check_model_validity,
)
from nebullvm.tools.base import (
    DeepLearningFramework,
    QuantizationType,
    Device,
    ModelCompiler,
)
from nebullvm.tools.utils import gpu_is_available


@pytest.mark.parametrize(
    (
        "output_library",
        "dynamic",
        "quantization_type",
        "metric_drop_ths",
        "metric",
    ),
    [
        (DeepLearningFramework.TENSORFLOW, False, None, None, None),
        (DeepLearningFramework.TENSORFLOW, True, None, None, None),
    ],
)
def test_tensorflow_backend(
    output_library: DeepLearningFramework,
    dynamic: bool,
    quantization_type: QuantizationType,
    metric_drop_ths: int,
    metric: str,
):
    device = Device.GPU if gpu_is_available() else Device.CPU
    with TemporaryDirectory() as tmp_dir:
        (
            model,
            input_data,
            model_params,
            input_tfms,
            model_outputs,
            metric,
        ) = initialize_model(dynamic, metric, output_library, device)

        compiler_op = TensorflowBackendCompiler()
        compiler_op.to(device).execute(
            model=model,
            input_tfms=input_tfms,
            metric_drop_ths=metric_drop_ths,
            quantization_type=quantization_type,
            input_data=input_data,
        )

        compiled_model = compiler_op.get_result()

        build_inference_learner_op = COMPILER_TO_INFERENCE_LEARNER_MAP[
            ModelCompiler.XLA
        ][DeepLearningFramework.TENSORFLOW]()

        build_inference_learner_op.to(device).execute(
            model=compiled_model,
            model_orig=compiler_op.model_orig
            if hasattr(compiler_op, "model_orig")
            else None,
            model_params=model_params,
            input_tfms=input_tfms,
            dl_framework=output_library,
        )

        optimized_model = build_inference_learner_op.get_result()

        assert isinstance(optimized_model, TensorflowBackendInferenceLearner)

        # Test save and load functions
        optimized_model.save(tmp_dir)
        loaded_model = TensorflowBackendInferenceLearner.load(tmp_dir)
        assert isinstance(loaded_model, TensorflowBackendInferenceLearner)

        assert isinstance(optimized_model.get_size(), int)

        inputs_example = list(optimized_model.get_inputs_example())
        res = optimized_model.predict(*inputs_example)
        assert res is not None

        # Test validity of the model
        valid = check_model_validity(
            optimized_model,
            input_data,
            model_outputs,
            metric_drop_ths,
            quantization_type,
            metric,
        )
        assert valid

        if dynamic:  # Check also with a smaller bath_size
            inputs_example = [
                input_[: len(input_) // 2] for input_ in inputs_example
            ]
            res = optimized_model.predict(*inputs_example)
            assert res is not None


@pytest.mark.parametrize(
    (
        "output_library",
        "dynamic",
        "quantization_type",
        "metric_drop_ths",
        "metric",
    ),
    [
        (
            DeepLearningFramework.TENSORFLOW,
            False,
            None,
            0.1,
            "numeric_precision",
        ),
        (
            DeepLearningFramework.TENSORFLOW,
            True,
            None,
            0.1,
            "numeric_precision",
        ),
        (
            DeepLearningFramework.TENSORFLOW,
            True,
            QuantizationType.DYNAMIC,
            2,
            "numeric_precision",
        ),
        (
            DeepLearningFramework.TENSORFLOW,
            True,
            QuantizationType.HALF,
            2,
            "numeric_precision",
        ),
        (
            DeepLearningFramework.TENSORFLOW,
            True,
            QuantizationType.STATIC,
            2,
            "numeric_precision",
        ),
    ],
)
def test_tf_lite(
    output_library: DeepLearningFramework,
    dynamic: bool,
    quantization_type: QuantizationType,
    metric_drop_ths: int,
    metric: str,
):
    device = Device.CPU
    with TemporaryDirectory() as tmp_dir:
        (
            model,
            input_data,
            model_params,
            input_tfms,
            model_outputs,
            metric,
        ) = initialize_model(dynamic, metric, output_library, device)

        compiler_op = TFLiteBackendCompiler()
        compiler_op.to(device).execute(
            model=model,
            input_tfms=input_tfms,
            metric_drop_ths=metric_drop_ths,
            quantization_type=quantization_type,
            input_data=input_data,
        )

        compiled_model = compiler_op.get_result()

        build_inference_learner_op = COMPILER_TO_INFERENCE_LEARNER_MAP[
            ModelCompiler.TFLITE
        ][DeepLearningFramework.TENSORFLOW]()

        build_inference_learner_op.to(device).execute(
            model=compiled_model,
            model_orig=compiler_op.model_orig
            if hasattr(compiler_op, "model_orig")
            else None,
            model_params=model_params,
            input_tfms=input_tfms,
            source_dl_framework=output_library,
        )

        optimized_model = build_inference_learner_op.get_result()

        assert isinstance(optimized_model, TFLiteBackendInferenceLearner)

        # Test save and load functions
        optimized_model.save(tmp_dir)
        loaded_model = TFLiteBackendInferenceLearner.load(tmp_dir)
        assert isinstance(loaded_model, TFLiteBackendInferenceLearner)

        assert isinstance(optimized_model.get_size(), int)

        inputs_example = list(optimized_model.get_inputs_example())
        res = optimized_model.predict(*inputs_example)
        assert res is not None

        # Test validity of the model
        valid = check_model_validity(
            optimized_model,
            input_data,
            model_outputs,
            metric_drop_ths,
            quantization_type,
            metric,
        )
        assert valid

        if dynamic:  # Check also with a smaller bath_size
            inputs_example = [
                input_[: len(input_) // 2] for input_ in inputs_example
            ]
            res = optimized_model.predict(*inputs_example)
            assert res is not None



from pathlib import Path

from loguru import logger

from nebullvm.config import ONNX_OPSET_VERSION
from nebullvm.optional_modules.torch import torch, Module
from nebullvm.tools.base import ModelParams, Device
from nebullvm.tools.data import DataManager
from nebullvm.tools.pytorch import (
    get_outputs_sizes_torch,
    create_model_inputs_torch,
)


def convert_torch_to_onnx(
    torch_model: Module,
    input_data: DataManager,
    model_params: ModelParams,
    output_file_path: Path,
    device: Device,
):
    """Function importing a custom model in pytorch and converting it in ONNX

    Args:
        torch_model (Module): Pytorch model.
        input_data (DataManager): Custom data provided by user to be
        used as input for the converter.
        model_params (ModelParams): Model Parameters as input sizes and
            dynamic axis information.
        output_file_path (str or Path): Path where storing the output
            ONNX file.
        device (Device): Device where the model will be run.
    """

    if input_data is not None:
        input_tensors = list(input_data.get_list(1)[0])
    else:
        input_tensors = create_model_inputs_torch(
            model_params.batch_size, model_params.input_infos
        )

    output_sizes = get_outputs_sizes_torch(torch_model, input_tensors, device)

    input_names = [f"input_{i}" for i in range(len(input_tensors))]
    output_names = [f"output_{i}" for i in range(len(output_sizes))]
    dynamic_info = model_params.dynamic_info
    if dynamic_info is not None:
        dynamic_info = {
            name: dynamic_dict
            for name, dynamic_dict in zip(
                input_names + output_names,
                dynamic_info.inputs + dynamic_info.outputs,
            )
        }

    try:
        # try conversion with model on gpu
        if device is Device.GPU:
            input_tensors = [x.cpu() for x in input_tensors]
            torch_model.cpu()

        torch.onnx.export(
            torch_model,  # model being run
            tuple(
                input_tensors
            ),  # model input (or a tuple for multiple inputs)
            str(output_file_path),
            # where to save the model (can be a file or file-like object)
            export_params=True,
            # store the trained parameter weights inside the model file
            opset_version=ONNX_OPSET_VERSION,
            # the ONNX version to export the model to
            do_constant_folding=True,
            # whether to execute constant folding for optimization
            input_names=input_names,
            # the model's input names
            output_names=output_names,
            # the model's output names
            dynamic_axes=dynamic_info,
        )

        # Put again model on gpu
        if device is Device.GPU:
            torch_model.cuda()

        return output_file_path
    except Exception:
        # try conversion with model on gpu
        if device is Device.GPU:
            input_tensors = [x.cuda() for x in input_tensors]
            torch_model.cuda()

            try:
                torch.onnx.export(
                    torch_model,  # model being run
                    tuple(
                        input_tensors
                    ),  # model input (or a tuple for multiple inputs)
                    str(output_file_path),
                    # where to save the model
                    # (can be a file or file-like object)
                    export_params=True,
                    # store the trained parameter weights inside the model file
                    opset_version=ONNX_OPSET_VERSION,
                    # the ONNX version to export the model to
                    do_constant_folding=True,
                    # whether to execute constant folding for optimization
                    input_names=input_names,
                    # the model's input names
                    output_names=output_names,
                    # the model's output names
                    dynamic_axes=dynamic_info,
                )

                return output_file_path
            except Exception:
                logger.warning(
                    "Exception raised during conversion from torch"
                    " to onnx model. ONNX pipeline will be unavailable."
                )
                return None
        else:
            logger.warning(
                "Exception raised during conversion from torch"
                " to onnx model. ONNX pipeline will be unavailable."
            )
            return None



import abc
import copy
import os
import subprocess
from pathlib import Path
from typing import List, Any, Tuple

import numpy as np

from nebullvm.config import QUANTIZATION_DATA_NUM, TORCH_TENSORRT_PRECISIONS
from nebullvm.operations.optimizations.compilers.base import Compiler

from nebullvm.operations.optimizations.compilers.quantizations.tensor_rt import (  # noqa: E501
    quantize_tensorrt,
)
from nebullvm.operations.optimizations.compilers.quantizations.utils import (
    check_quantization,
)
from nebullvm.optional_modules.tensor_rt import tensorrt as trt
from nebullvm.optional_modules.torch import torch, Module
from nebullvm.optional_modules.torch_tensorrt import (
    torch_tensorrt,
    DataLoaderCalibrator,
)
from nebullvm.tools.base import (
    QuantizationType,
    ModelParams,
)
from nebullvm.tools.data import DataManager, PytorchDataset
from nebullvm.tools.onnx import get_input_names
from nebullvm.tools.transformations import (
    MultiStageTransformation,
    HalfPrecisionTransformation,
)


class TensorRTCompiler(Compiler, abc.ABC):
    supported_ops = {
        "cpu": [],
        "gpu": [
            None,
            QuantizationType.STATIC,
            QuantizationType.HALF,
        ],
    }

    def __init__(self):
        super().__init__()
        self.model_orig = None

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        pass


class PyTorchTensorRTCompiler(TensorRTCompiler):
    def execute(
        self,
        model: Module,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Compile the input model using TensorRT Compiler from the
            PyTorch interface.

        Args:
            model (torch.nn.Module): The pytorch model.
            model_params (ModelParams): The model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction. Default: None.
            metric_drop_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with a higher drop
                will be ignored. Default: None.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used. Default: None.
            input_data (DataManager): User defined data. Default: None
        """

        if quantization_type not in self.supported_ops[self.device.value]:
            self.compiled_model = None
            return

        if quantization_type is QuantizationType.STATIC and input_data is None:
            raise ValueError("Input data is required for static quantization.")

        self.logger.info(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )

        check_quantization(quantization_type, metric_drop_ths)

        if quantization_type is QuantizationType.HALF:
            dtype = torch.half
            input_tfms.append(HalfPrecisionTransformation())
        elif quantization_type is QuantizationType.STATIC:
            dtype = torch.int8

            dataset = PytorchDataset(input_data.get_split("train"))
            dataloader = torch.utils.data.DataLoader(
                dataset,
                batch_size=dataset.batch_size,
                shuffle=False,
                num_workers=0,
            )

            calibrator = torch_tensorrt.ptq.DataLoaderCalibrator(
                dataloader,
                use_cache=False,
                algo_type=torch_tensorrt.ptq.CalibrationAlgo.ENTROPY_CALIBRATION_2,  # noqa E501
                device=torch.device("cuda:0"),
            )
        else:
            dtype = torch.float32

        # Convert int64 to int32 for transformers inputs
        input_tensors = [
            tensor.cuda()
            if tensor.dtype != torch.int64
            else tensor.to(torch.int32).cuda()
            for tensor in input_data.get_list(1)[0]
        ]

        self.compiled_model = self._compile_model(
            model=model,
            input_tensors=input_tensors,
            dtype=dtype,
            calibrator=calibrator
            if quantization_type is QuantizationType.STATIC
            else None,  # noqa E501
            quantization_type=quantization_type,
        )

    def _compile_model(
        self,
        model: Module,
        input_tensors: List[torch.Tensor],
        dtype: torch.dtype,
        calibrator: DataLoaderCalibrator,
        quantization_type: QuantizationType,
    ):

        model.cuda().eval()

        try:
            torch.jit.script(model)
        except Exception:
            model = torch.jit.trace(model, input_tensors)

        with torch_tensorrt.logging.errors():
            trt_model = torch_tensorrt.compile(
                model
                if dtype is not torch.half
                else copy.deepcopy(model).half(),
                inputs=[
                    torch_tensorrt.Input(
                        tensor.shape,
                        dtype=torch.half
                        if (
                            dtype == torch.half
                            and tensor.dtype not in [torch.int8, torch.int32]
                        )
                        else tensor.dtype,
                    )
                    for tensor in input_tensors
                ],
                enabled_precisions=TORCH_TENSORRT_PRECISIONS[str(dtype)],
                calibrator=calibrator
                if quantization_type is QuantizationType.STATIC
                else None,
                workspace_size=1 << 30,
                device={
                    "device_type": torch_tensorrt.DeviceType.GPU,
                    "gpu_id": 0,
                    "dla_core": 0,
                    "allow_gpu_fallback": False,
                    "disable_tf32": False,
                },
                truncate_long_and_double=True,
            )

        # Delete calibration cache
        if os.path.exists("calibration.cache"):
            os.remove("calibration.cache")

        return trt_model

    @staticmethod
    def _quantize_model(**kwargs) -> Any:
        raise NotImplementedError


class ONNXTensorRTCompiler(TensorRTCompiler):
    def __init__(self):
        super().__init__()
        self.model_orig = None

    def execute(
        self,
        model: str,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Compile the input model using TensorRT Compiler from the
            ONNX interface.

        Args:
            model (str): The path to the onnx model.
            model_params (ModelParams): The model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction. Default: None.
            metric_drop_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with a higher drop
                will be ignored. Default: None.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used. Default: None.
            input_data (DataManager): User defined data. Default: None
        """

        if quantization_type not in self.supported_ops[self.device.value]:
            self.compiled_model = None
            return

        if quantization_type is QuantizationType.STATIC and input_data is None:
            raise ValueError("Input data is required for static quantization.")

        self.logger.info(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )

        check_quantization(quantization_type, metric_drop_ths)
        train_input_data = input_data.get_split("train").get_numpy_list(
            QUANTIZATION_DATA_NUM
        )

        try:
            import onnxsim  # noqa: F401

            # Simplify model, otherwise tensor RT won't work on gpt2 and some
            # other models.
            simplified_model = str(model) + "_simplified"
            if not Path(simplified_model).is_file():
                cmd = [
                    "onnxsim",
                    str(model),
                    simplified_model,
                ]
                subprocess.run(cmd, stdout=subprocess.DEVNULL)

            # First try with simplified model
            onnx_model_path = simplified_model
        except Exception:
            # Try again with original model
            onnx_model_path = str(model)

        # -- Build phase --
        nvidia_logger = trt.Logger(trt.Logger.ERROR)
        builder = trt.Builder(nvidia_logger)
        # create network definition
        network = builder.create_network(
            1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
        )
        # build the engine
        # TODO: setup config value for the class in a config file
        config = builder.create_builder_config()
        try:
            config.set_memory_pool_limit(trt.MemoryPoolType.WORKSPACE, 1 << 30)
        except AttributeError:
            # The method set_memory_pool_limit is not available
            # until TensorRT Release 8.4.1
            self.logger.warning(
                "Cannot call method set_memory_pool_limit for TensorRT."
                "Please update TensorRT version."
            )

        if quantization_type is not None:
            config = self._quantize_model(
                quantization_type,
                model_params,
                config,
                input_tfms,
                train_input_data
                if quantization_type is QuantizationType.STATIC
                else None,
            )

        self.compiled_model = self._compile_model(
            onnx_model_path=str(onnx_model_path),
            model_params=model_params,
            config=config,
            network=network,
            builder=builder,
            nvidia_logger=nvidia_logger,
        )
        self.model_orig = onnx_model_path

    def _compile_model(
        self,
        onnx_model_path: str,
        model_params: ModelParams,
        config,
        network,
        builder,
        nvidia_logger,
    ):
        parser = trt.OnnxParser(network, nvidia_logger)
        success = parser.parse_from_file(onnx_model_path)

        if not success:
            for idx in range(parser.num_errors):
                self.logger.debug(parser.get_error(idx))
            raise ValueError(
                f"Errors occurred while processing the "
                f"ONNX file at {onnx_model_path}"
            )

        if model_params.dynamic_info is not None:
            profile = builder.create_optimization_profile()
            for input_name, input_dynamic_info, input_info in zip(
                get_input_names(onnx_model_path),
                model_params.dynamic_info.inputs,
                model_params.input_infos,
            ):
                profile.set_shape(
                    input_name,
                    (
                        min(model_params.batch_size, 1)
                        if 0 in input_dynamic_info
                        else model_params.batch_size,
                        *(
                            shape
                            if i + 1 not in input_dynamic_info
                            else (input_info.min_sizes or {}).get(i + 1, 1)
                            for i, shape in enumerate(input_info.size)
                        ),
                    ),
                    (model_params.batch_size, *input_info.size),
                    (model_params.batch_size, *input_info.size),
                )
            config.add_optimization_profile(profile)
        return builder.build_serialized_network(network, config)

    @staticmethod
    def _quantize_model(
        quantization_type: QuantizationType,
        model_params: ModelParams,
        config,
        input_tfms: MultiStageTransformation,
        input_data: List[Tuple[np.ndarray, ...]] = None,
    ):
        return quantize_tensorrt(
            quantization_type,
            model_params,
            config,
            input_tfms,
            input_data,
        )



try:
    import tvm
    from tvm import IRModule
    from tvm.runtime.ndarray import NDArray
    from tvm.autotvm.tuner import XGBTuner
    from tvm import autotvm
    import tvm.relay as relay
    from tvm.relay.transform import ToMixedPrecision
    from tvm.contrib.graph_executor import GraphModule
    from tvm.runtime import Module
    from tvm.relay.backend.executor_factory import ExecutorFactoryModule
except ImportError:
    tvm = (
        IRModule
    ) = (
        NDArray
    ) = (
        XGBTuner
    ) = (
        ExecutorFactoryModule
    ) = autotvm = relay = ToMixedPrecision = GraphModule = Module = object



from nebullvm.tools.base import ModelCompiler, Device


def onnxruntime_is_available() -> bool:
    try:
        import onnxruntime  # noqa F401

        return True
    except ImportError:
        return False


def tvm_is_available() -> bool:
    try:
        import tvm  # noqa F401

        return True
    except ImportError:
        return False


def bladedisc_is_available() -> bool:
    try:
        import torch_blade  # noqa F401

        return True
    except ImportError:
        return False


def tensorrt_is_available() -> bool:
    try:
        import tensorrt  # noqa F401
        import polygraphy  # noqa F401

        return True
    except ImportError:
        return False


def torch_tensorrt_is_available() -> bool:
    try:
        import torch_tensorrt  # noqa F401

        return True
    except ImportError:
        return False


def openvino_is_available() -> bool:
    try:
        from openvino.runtime import Core  # noqa F401
    except ImportError:
        return False
    else:
        return True


def deepsparse_is_available() -> bool:
    try:
        import deepsparse  # noqa F401
    except ImportError:
        return False
    else:
        return True


def intel_neural_compressor_is_available() -> bool:
    try:
        import neural_compressor  # noqa F401
    except ImportError:
        return False
    else:
        return True


def select_compilers_from_hardware_onnx(device: Device):
    from nebullvm.optional_modules.utils import onnx_is_available

    compilers = []
    if onnx_is_available():
        if onnxruntime_is_available():
            compilers.append(ModelCompiler.ONNX_RUNTIME)
        if tvm_is_available():
            compilers.append(ModelCompiler.APACHE_TVM)
        if device is Device.GPU and tensorrt_is_available():
            compilers.append(ModelCompiler.TENSOR_RT)
        if device is Device.CPU and openvino_is_available():
            compilers.append(ModelCompiler.OPENVINO)
    return compilers


def select_compilers_from_hardware_torch(device: Device):
    from nebullvm.optional_modules.utils import torch_is_available

    compilers = []
    if torch_is_available():
        compilers.append(ModelCompiler.TORCHSCRIPT)
        if tvm_is_available():
            compilers.append(ModelCompiler.APACHE_TVM)
        if bladedisc_is_available():
            compilers.append(ModelCompiler.BLADEDISC)

        if device is Device.CPU:
            if deepsparse_is_available():
                compilers.append(ModelCompiler.DEEPSPARSE)
            if intel_neural_compressor_is_available():
                compilers.append(ModelCompiler.INTEL_NEURAL_COMPRESSOR)
        elif device is Device.GPU:
            if torch_tensorrt_is_available:
                compilers.append(ModelCompiler.TENSOR_RT)
    return compilers


def select_compilers_from_hardware_tensorflow():
    from nebullvm.optional_modules.utils import tensorflow_is_available

    compilers = []
    if tensorflow_is_available():
        compilers.append(ModelCompiler.XLA)
        compilers.append(ModelCompiler.TFLITE)
    return compilers



import abc
from typing import Tuple, Optional

from nebullvm.operations.base import Operation


class Measure(Operation, abc.ABC):
    def __init__(self):
        super().__init__()
        self.measure_result = None

    @abc.abstractmethod
    def execute(self, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_result(self) -> Optional[Tuple]:
        raise NotImplementedError()



from abc import ABC, abstractmethod
from typing import Any, Optional, Dict, Callable, Tuple

import yaml

from nebullvm.operations.base import Operation
from nebullvm.tools.data import DataManager


class Compressor(Operation, ABC):
    def __init__(self, config_file: str = None):
        super().__init__()
        self._config = self._read_config(config_file)
        self.compressed_model = None
        self.new_metric_ths = None

    @abstractmethod
    def execute(
        self,
        model: Any,
        train_input_data: DataManager,
        eval_input_data: DataManager,
        metric_drop_ths: float,
        metric: Callable,
    ) -> Tuple[Any, Optional[float]]:
        raise NotImplementedError()

    def _read_config(self, config_file: Optional[str]) -> Dict:
        config = self._get_default_config()
        if config_file is not None:
            with open(config_file, "r") as f:
                data = yaml.load(f, Loader=yaml.CLoader)
                config.update(data.get(self.config_key, {}))
        return config

    @staticmethod
    @abstractmethod
    def _get_default_config() -> Dict:
        raise NotImplementedError

    @property
    @abstractmethod
    def config_key(self) -> str:
        raise NotImplementedError()

    def get_result(self) -> Tuple[Any, Optional[float]]:
        return self.compressed_model, self.new_metric_ths



try:
    from transformers import (
        PreTrainedModel,
    )
    from transformers.tokenization_utils import PreTrainedTokenizer
except ImportError:
    # add placeholders for function definition
    PreTrainedModel = None
    PreTrainedTokenizer = None



import json
import logging
import os.path
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Tuple, List, Any, Dict

import torch
from sparseml.onnx.optim import ModelAnalyzer, pruning_loss_sens_magnitude
from sparseml.pytorch.optim import (
    ScheduledModifierManager,
)
from sparseml.pytorch.sparsification import (
    EpochRangeModifier,
    GMPruningModifier,
)
from sparseml.pytorch.utils import ModuleExporter
from sparsify.blueprints.utils import (
    default_epochs_distribution,
    PruningModelEvaluator,
    default_pruning_settings,
)
from sparsify.schemas import ProjectModelAnalysisSchema
from torch.nn import CrossEntropyLoss, MSELoss
from torch.optim import SGD
from tqdm.auto import tqdm

CRITERION_FNS = {
    "CrossEntropy": CrossEntropyLoss(),
    "MSE": MSELoss(),
}

logging.basicConfig(
    format=" %(asctime)s [%(levelname)s] %(message)s",
    datefmt="%d/%m/%Y %I:%M:%S %p",
)
logger = logging.getLogger("nebullvm_logger")
logger.setLevel(logging.INFO)


def _export_model_onnx(
    model: torch.nn.Module,
    save_path: Path,
    model_name: str,
    input_batch: Tuple,
):
    if torch.cuda.is_available():
        input_batch = tuple(t.cuda() for t in input_batch)
        model.cuda()

    exporter = ModuleExporter(model, output_dir=save_path)
    with torch.no_grad():
        example_outputs = model(*input_batch)
    exporter.export_onnx(
        input_batch, name=model_name, example_outputs=example_outputs
    )
    onnx_path = save_path / model_name

    return onnx_path


class RecipeBuilder:
    def __init__(self, model_path):
        self.model_path = model_path

    def _make_analysis(self):
        analyzer = ModelAnalyzer(self.model_path)
        self.analysis = ProjectModelAnalysisSchema().load(analyzer.dict())

    def _compute_loss_sensitivity(self):
        sensitivities = []
        parameters = []
        for i, node in enumerate(self.analysis["nodes"]):
            if node["prunable"]:
                sensitivities.append(node["prunable_equation_sensitivity"])
                parameters.append(node["prunable_params"])

        loss_analysis = pruning_loss_sens_magnitude(self.model_path)

        results_model = loss_analysis.results_model
        results = loss_analysis.results

        model = {
            "baseline_measurement_key": (
                str(results_model.baseline_measurement_key)
            ),
            "measurements": {
                str(key): val for key, val in results_model.averages.items()
            },
        }
        ops = []

        for res in results:
            ops.append(
                {
                    "id": res.id_,
                    "name": res.name,
                    "index": res.index,
                    "baseline_measurement_key": (
                        str(res.baseline_measurement_key)
                    ),
                    "measurements": {
                        str(key): val for key, val in res.averages.items()
                    },
                }
            )

        pruning = {"model": model, "ops": ops}
        loss = {}
        loss["baseline"] = {}
        loss["pruning"] = pruning

        model = PruningModelEvaluator(
            self.analysis,
            None,
            loss,
        )
        model.eval_baseline(default_pruning_settings().sparsity)
        model.eval_pruning(default_pruning_settings())

        self.final_analysis = model.to_dict_values()

    def build_recipe(self, epochs_pruning_window=None, training_epochs=10):
        self._make_analysis()
        self._compute_loss_sensitivity()

        if epochs_pruning_window is None:
            epochs = default_epochs_distribution(training_epochs)
        else:
            # TODO: set custom parameters
            epochs = default_epochs_distribution(training_epochs)
            epochs_dict = epochs._asdict()
            epochs_dict.update(epochs_pruning_window)
            epochs = epochs.__class__(**epochs_dict)

        mods = [
            EpochRangeModifier(
                start_epoch=epochs.start_epoch,
                end_epoch=epochs.end_epoch,
            )
        ]

        node_weight_name_lookup = {
            node["id"]: node["weight_name"]
            for node in self.analysis["nodes"]
            if node["prunable"]
        }

        sparsity_to_params = {}

        nodes = self.final_analysis[0]

        for node in nodes:
            sparsity = node["sparsity"]
            node_id = node["node_id"]
            weight_name = node_weight_name_lookup[node_id]

            if sparsity is None:
                continue

            if sparsity not in sparsity_to_params:
                sparsity_to_params[sparsity] = []

            sparsity_to_params[sparsity].append(weight_name)

        for sparsity, params in sparsity_to_params.items():
            gm_pruning = GMPruningModifier(
                init_sparsity=0.05,
                final_sparsity=sparsity,
                start_epoch=epochs.pruning_start_epoch,
                end_epoch=epochs.pruning_end_epoch,
                update_frequency=epochs.pruning_update_frequency,
                params=params,
            )

            mods.append(gm_pruning)

        return ScheduledModifierManager(mods)


class PruningTrainer:
    def __init__(self, model, bs):
        self.data_loader = None
        self.optimizer = None
        self.model = model
        self.batch_size = bs

    def _setup_training(self, loss_fn=None, lr=1e-3, momentum=0.9):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model.to(self.device)
        if loss_fn is None:
            loss_fn = CrossEntropyLoss()
        else:
            loss_fn = CRITERION_FNS.get(loss_fn, CrossEntropyLoss())
        self.criterion = loss_fn
        self.optimizer = SGD(self.model.parameters(), lr=lr, momentum=momentum)

    def _run_model_one_epoch(self, train=False):

        if train:
            self.model.train()
            data_loader = self.train_data_loader
        else:
            self.model.eval()
            data_loader = self.val_data_loader

        running_loss = 0.0

        for step, (inputs, labels) in tqdm(
            enumerate(data_loader), total=len(data_loader)
        ):
            inputs = tuple(t.to(self.device) for t in inputs)
            if not isinstance(labels, torch.Tensor):
                labels = torch.tensor(labels)
                if len(labels.shape) == 0:
                    labels = labels.unsqueeze(0)
            labels = labels.to(self.device)

            if train:
                self.optimizer.zero_grad()

            outputs = self.model(
                *inputs
            )  # model returns logits and softmax as a tuple
            loss = self.criterion(outputs, labels)

            if train:
                loss.backward()
                self.optimizer.step()

            running_loss += loss.item()

        loss = running_loss / (len(data_loader) + 1e-5)
        return loss

    def train(
        self, manager, train_data_loader, val_data_loader, **train_kwargs
    ):
        self.train_data_loader = train_data_loader
        self.val_data_loader = val_data_loader
        self._setup_training(**train_kwargs)
        self.optimizer = manager.modify(
            self.model,
            self.optimizer,
            steps_per_epoch=len(self.train_data_loader),
        )
        self.model.train()
        # Run model pruning
        epoch = manager.min_epochs
        while epoch < manager.max_epochs:
            # run training loop
            epoch_name = "{}/{}".format(epoch + 1, manager.max_epochs)
            logger.info("Running Training Epoch {}".format(epoch_name))
            train_loss = self._run_model_one_epoch(train=True)
            logger.info(
                ("Training Epoch: {}\nTraining Loss: {}\n").format(
                    epoch_name, train_loss
                )
            )

            # run validation loop
            logger.info("Running Validation Epoch {}".format(epoch_name))
            val_loss = self._run_model_one_epoch()
            logger.info(
                "Validation Epoch: {}\nVal Loss: {}\n".format(
                    epoch_name, val_loss
                )
            )

            epoch += 1

        manager.finalize(self.model)

        return self.model


def _load_config(config_file: str):
    with open(config_file, "r") as f:
        config = json.load(f)
    return config


def _load_data(data_dir: str):
    data_dir = Path(data_dir)
    return [torch.load(input_path) for input_path in data_dir.glob("*.pt")]


def _load_model(model_file: str):
    if os.path.isdir(model_file):
        path = Path(model_file)
        module_file = path / "module.py"
        with open(module_file, "r") as f:
            module_str = f.read()
        exec(module_str, globals())
        model = eval("NebullvmFxModule")()
        model.load_state_dict(torch.load(path / "state_dict.pt"))
    else:
        model = torch.load(model_file)
    return model


def _train_model(
    model: torch.nn.Module,
    train_data: List[Tuple[Tuple, Any]],
    eval_data: List[Tuple[Tuple, Any]],
    epochs_pruning_window: Dict = None,
    training_epochs: int = 10,
    lr: float = 1e-3,
    momentum: float = 0.9,
    loss_fn: str = "CrossEntropy",
):
    batch_size = train_data[0][0][0].shape[0]
    with TemporaryDirectory() as tmp_dir:
        onnx_path = _export_model_onnx(
            model, Path(tmp_dir), "model.onnx", train_data[0][0]
        )
        onnx_path = onnx_path.as_posix()

        recipe = RecipeBuilder(onnx_path)
        # TODO: implement custom parameters support
        manager = recipe.build_recipe(
            epochs_pruning_window=epochs_pruning_window,
            training_epochs=training_epochs,
        )
        trainer = PruningTrainer(model, batch_size)
        pruned_model = trainer.train(
            manager, train_data, eval_data, lr=lr, momentum=momentum
        )
        return pruned_model


def _save_model(model: torch.nn.Module, path: str):
    if path.endswith(".pt"):
        torch.save(model, path)
    else:
        torch.save(model.state_dict(), Path(path) / "pruned_state_dict.pt")


def main(
    model_file: str,
    train_data_dir: str,
    eval_data_dir: str,
    config_file: str,
    out_file: str,
):
    config = _load_config(config_file)
    model = _load_model(model_file)
    train_data = _load_data(train_data_dir)
    eval_data = _load_data(eval_data_dir)
    pruned_model = _train_model(model, train_data, eval_data, **config)
    _save_model(pruned_model, out_file)


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("--model", help="The model to be pruned.")
    parser.add_argument(
        "--train_dir",
        help="The directory contained the pickled training data.",
    )
    parser.add_argument(
        "--eval_dir", help="The directory contained the pickled test data."
    )
    parser.add_argument("--config", help="The config file.")
    parser.add_argument(
        "--pruned_model", help="Path where storing the pruned model."
    )
    args = parser.parse_args()
    main(
        model_file=args.model,
        train_data_dir=args.train_dir,
        eval_data_dir=args.eval_dir,
        config_file=args.config,
        out_file=args.pruned_model,
    )



from pathlib import Path
from tempfile import TemporaryDirectory

import cpuinfo
import pytest

from nebullvm.operations.conversions.converters import PytorchConverter
from nebullvm.operations.inference_learners.openvino import (
    OPENVINO_INFERENCE_LEARNERS,
)
from nebullvm.operations.optimizations.base import (
    COMPILER_TO_INFERENCE_LEARNER_MAP,
)
from nebullvm.operations.optimizations.compilers.openvino import (
    OpenVINOCompiler,
)
from nebullvm.operations.optimizations.tests.utils import (
    initialize_model,
    check_model_validity,
)
from nebullvm.tools.base import (
    Device,
    DeepLearningFramework,
    QuantizationType,
    ModelCompiler,
)


@pytest.mark.parametrize(
    (
        "output_library",
        "dynamic",
        "quantization_type",
        "metric_drop_ths",
        "metric",
    ),
    [
        (DeepLearningFramework.PYTORCH, True, None, None, None),
        (DeepLearningFramework.PYTORCH, False, None, None, None),
        (
            DeepLearningFramework.PYTORCH,
            False,
            QuantizationType.HALF,
            2,
            "numeric_precision",
        ),
        (
            DeepLearningFramework.PYTORCH,
            False,
            QuantizationType.STATIC,
            2,
            "numeric_precision",
        ),
    ],
)
@pytest.mark.skipif(
    "intel" not in cpuinfo.get_cpu_info()["brand_raw"].lower(),
    reason="Openvino is only available for intel processors.",
)
def test_openvino(
    output_library: DeepLearningFramework,
    dynamic: bool,
    quantization_type: QuantizationType,
    metric_drop_ths: int,
    metric: str,
):
    device = Device.CPU
    with TemporaryDirectory() as tmp_dir:
        (
            model,
            input_data,
            model_params,
            input_tfms,
            model_outputs,
            metric,
        ) = initialize_model(dynamic, metric, output_library, device)

        model_path = Path(tmp_dir) / "fp32"
        model_path.mkdir(parents=True)

        converter_op = PytorchConverter()
        converter_op.to(device).set_state(model, input_data).execute(
            model_path, model_params
        )

        converted_models = converter_op.get_result()
        assert len(converted_models) > 1

        model_path = str(
            [model for model in converted_models if isinstance(model, Path)][0]
        )

        compiler_op = OpenVINOCompiler()
        compiler_op.to(device).execute(
            model=model_path,
            model_params=model_params,
            input_tfms=input_tfms,
            metric_drop_ths=metric_drop_ths,
            quantization_type=quantization_type,
            input_data=input_data,
        )

        compiled_model = compiler_op.get_result()

        build_inference_learner_op = COMPILER_TO_INFERENCE_LEARNER_MAP[
            ModelCompiler.OPENVINO
        ][DeepLearningFramework.NUMPY]()
        build_inference_learner_op.to(device).execute(
            model=compiled_model,
            model_orig=compiler_op.model_orig
            if hasattr(compiler_op, "model_orig")
            else None,
            model_params=model_params,
            input_tfms=input_tfms,
            source_dl_framework=output_library,
        )

        optimized_model = build_inference_learner_op.get_result()
        assert isinstance(
            optimized_model, OPENVINO_INFERENCE_LEARNERS[output_library]
        )

        # Test save and load functions
        optimized_model.save(tmp_dir)
        loaded_model = OPENVINO_INFERENCE_LEARNERS[output_library].load(
            tmp_dir
        )
        assert isinstance(
            loaded_model, OPENVINO_INFERENCE_LEARNERS[output_library]
        )

        assert isinstance(optimized_model.get_size(), int)

        inputs_example = list(optimized_model.get_inputs_example())
        res = optimized_model(*inputs_example)
        assert res is not None

        # Test validity of the model
        valid = check_model_validity(
            optimized_model,
            input_data,
            model_outputs,
            metric_drop_ths,
            quantization_type,
            metric,
        )
        assert valid

        if dynamic:  # Check also with a smaller bath_size
            inputs_example = [
                input_[: len(input_) // 2] for input_ in inputs_example
            ]
            res = optimized_model(*inputs_example)
            assert res is not None



try:
    import onnx  # noqa F401
except ImportError:
    onnx = None

try:
    import onnxmltools  # noqa F401
    from onnxmltools.utils.float16_converter import (  # noqa F401
        convert_float_to_float16_model_path,
    )

except ImportError:
    convert_float_to_float16_model_path = object
